# -*- coding: utf-8 -*-
"""Classes that represent a ‘topocentric’ position on the Earth’s surface."""

from numpy import arctan2, array, array2string, cos, exp, sin, sqrt
from .constants import ANGVEL, DAY_S, RAD2DEG, T0, pi, tau
from .earthlib import refract
from .framelib import itrs
from .functions import (
    _T, angular_velocity_matrix, mxm, mxv, rot_y, rot_z,
)
from .descriptorlib import reify
from .units import Angle, Distance, _ltude
from .vectorlib import VectorFunction

_EARTH_ANGULAR_VELOCITY_VECTOR = array((0, 0, DAY_S * ANGVEL))

_lat_options = {'precision': 4, 'floatmode': 'fixed', 'sign': '+',
                'threshold': 5, 'edgeitems': 2}
_lon_options = {'precision': 4, 'floatmode': 'fixed',
                'threshold': 5, 'edgeitems': 2}
_elev_options = {'precision': 1, 'floatmode': 'fixed',
                 'threshold': 5, 'edgeitems': 2}

class ITRSPosition(VectorFunction):
    """An |xyz| position in the Earth-centered Earth-fixed (ECEF) ITRS frame."""

    center = 399

    def __init__(self, itrs_xyz):
        self.itrs_xyz = itrs_xyz
        x, y, z = itrs_xyz.au
        self._velocity_au_per_d = ANGVEL * DAY_S * array((-y, x, 0.0 * z))

    @property
    def target(self):
        # When used as a vector function, this Earth geographic location
        # computes positions from the Earth's center to itself.  (This
        # is a property, rather than an attribute, to avoid a circular
        # reference that delays garbage collection.)
        return self

    def _at(self, t):
        """Compute GCRS position and velocity at time `t`."""
        r = self.itrs_xyz.au
        v = self._velocity_au_per_d

        RT = _T(itrs.rotation_at(t))
        r = mxv(RT, r)
        v = mxv(RT, v)
        return r, v, None, None

class GeographicPosition(ITRSPosition):
    """A latitude-longitude-elevation position on Earth.

    Each instance of this class holds an |xyz| vector for a geographic
    position on, above, or below the Earth’s surface, in the ITRS
    reference frame: the international standard for an Earth-centered
    Earth-fixed (ECEF) reference frame.  Instead of instantiating this
    class directly, Skyfield users usually give a reference geoid the
    longitude and latitude they are interested in::

        from skyfield.api import wgs84
        topos = wgs84.latlon(37.3414, -121.6429)

    Once a geographic position has been created, here are its attributes
    and methods:

    """
    vector_name = 'Geodetic'

    def __init__(self, model, latitude, longitude, elevation, itrs_xyz):
        super(GeographicPosition, self).__init__(itrs_xyz)
        self.model = model
        self.latitude = latitude
        self.longitude = longitude
        self.elevation = elevation
        self._R_lat = _R_lat = rot_y(latitude.radians)[::-1]
        self._R_latlon = mxm(_R_lat, rot_z(-longitude.radians))

    @reify  # not @property, so users have the option of overwriting it
    def target_name(self):
        return '{0} latitude {1} N longitude {2} E elevation {3} m'.format(
            self.model.name,
            array2string(self.latitude.degrees, **_lat_options),
            array2string(self.longitude.degrees, **_lon_options),
            array2string(self.elevation.m, **_elev_options))

    def lst_hours_at(self, t):
        """Return the Local Apparent Sidereal Time, in hours, at time ``t``.

        This location’s Local Apparent Sidereal Time (LAST) is the right
        ascension of the zenith at the time ``t``, as measured against
        the “true” Earth equator and equinox (rather than the fictional
        “mean” equator and equinox, which ignore the Earth’s nutation).

        """
        sprime = -47.0e-6 * (t.whole - T0 + t.tdb_fraction) / 36525.0
        return (t.gast + self.longitude.hours + sprime / 54000.0) % 24.0

    def refract(self, altitude_degrees, temperature_C, pressure_mbar):
        """Predict how the atmosphere will refract a position.

        Given a body that is standing ``altitude_degrees`` above the
        true horizon, return an ``Angle`` predicting its apparent
        altitude given the supplied temperature and pressure, either of
        which can be the string ``'standard'`` to use 10°C and a
        pressure of 1010 mbar adjusted for the elevation of this
        geographic location.

        """
        if temperature_C == 'standard':
            temperature_C = 10.0
        if pressure_mbar == 'standard':
            pressure_mbar = 1010.0 * exp(-self.elevation.m / 9.1e3)
        alt = refract(altitude_degrees, temperature_C, pressure_mbar)
        return Angle(degrees=alt)

    def rotation_at(self, t):
        """Compute rotation from GCRS to this location’s altazimuth system."""
        return mxm(self._R_latlon, itrs.rotation_at(t))

    def _dRdt_times_RT_at(self, t):
        # TODO: taking the derivative of the instantaneous angular
        # velocity would provide a more accurate transform.
        R = mxv(self._R_lat, _EARTH_ANGULAR_VELOCITY_VECTOR)
        return angular_velocity_matrix(R)

class Geoid(object):
    """An Earth ellipsoid: maps latitudes and longitudes to |xyz| positions.

    Instead of creating their own geoid object, most Skyfield users
    simply use the `wgs84` object that comes built-in.

    The math for turning a position into latitude and longitude is based
    on Dr. T.S. Kelso's quite helpful article `Orbital Coordinate
    Systems, Part III <https://www.celestrak.org/columns/v02n03/>`_.

    """
    def __init__(self, name, radius_m, inverse_flattening):
        self.name = name
        self.radius = Distance(m=radius_m)
        self.inverse_flattening = inverse_flattening
        omf = (inverse_flattening - 1.0) / inverse_flattening
        self._one_minus_flattening_squared = omf * omf
        f = 1.0 / inverse_flattening
        self._e2 = 2.0*f - f*f

    @reify
    def polar_radius(self):
        """The Earth’s polar radius, as a :class:`~skyfield.units.Distance`."""
        return Distance(self.radius.au * (1.0 - 1.0 / self.inverse_flattening))

    def latlon(self, latitude_degrees, longitude_degrees, elevation_m=0.0,
               cls=GeographicPosition):
        """Return a `GeographicPosition` for a given latitude and longitude.

        The longitude and latitude should both be specified in degrees.
        If no elevation in meters is supplied, the returned position
        will lie on the surface of the ellipsoid.  Longitude is positive
        towards the east, so supply a negative number for west::

            from skyfield.api import wgs84
            observatory = wgs84.latlon(37.3414, -121.6429)  # 121.6° West

        You can avoid remembering which directions are negative by using
        Skyfield’s compass direction constants, which have the values +1
        and −1::

            from skyfield.api import N, S, E, W
            observatory = wgs84.latlon(37.3414 * N, 121.6429 * W)

        """
        latitude = Angle(degrees=latitude_degrees)
        longitude = Angle(degrees=longitude_degrees)
        elevation = Distance(m=elevation_m)

        lat = latitude.radians
        lon = longitude.radians
        radius_au = self.radius.au
        elevation_au = elevation.au

        sinphi = sin(lat)
        cosphi = cos(lat)
        omf2 = self._one_minus_flattening_squared
        c = 1.0 / sqrt(cosphi * cosphi + sinphi * sinphi * omf2)
        s = omf2 * c

        # At equator: 6378 km, the Earth's actual radius at the equator.
        # At the pole: 6399 km, the Earth's radius of curvature at the pole.
        radius_xy = radius_au * c
        xy = (radius_xy + elevation_au) * cosphi
        x = xy * cos(lon)
        y = xy * sin(lon)

        # At equator: 6335 km, the Earth's radius of curvature at the equator.
        # At the pole: 6357 km, the Earth's actual radius at the pole.
        radius_z = radius_au * s
        z = (radius_z + elevation_au) * sinphi

        r = array((x, y, z))
        return cls(self, latitude, longitude, elevation, Distance(r))

    def latlon_of(self, position):
        """Return the latitude and longitude of a ``position``.

        The position’s ``.center`` must be 399, the center of the Earth.
        Geodetic latitude and longitude are returned as a pair of
        :class:`~skyfield.units.Angle` objects.

        """
        xyz_au, x, y, R, aC, hyp, lat = self._compute_latitude(position)
        lon = (arctan2(y, x) - pi) % tau - pi
        return Angle(radians=lat), Angle(radians=lon)

    def height_of(self, position):
        """Return the height above the Earth’s ellipsoid of a ``position``.

        The position’s ``.center`` must be 399, the center of the Earth.
        A :class:`~skyfield.units.Distance` is returned giving the
        position’s geodetic height above the Earth’s surface.

        """
        xyz_au, x, y, R, aC, hyp, lat = self._compute_latitude(position)
        height_au = sqrt(hyp * hyp + R * R) - aC
        return Distance(height_au)

    def geographic_position_of(self, position):
        """Return the `GeographicPosition` of a ``position``.

        The position’s ``.center`` must be 399, the center of the Earth.
        A `GeographicPosition` is returned giving the position’s
        geodetic ``latitude`` and ``longitude``, and an ``elevation``
        above or below the surface of the ellipsoid.

        """
        xyz_au, x, y, R, aC, hyp, lat = self._compute_latitude(position)
        lon = (arctan2(y, x) - pi) % tau - pi
        height_au = sqrt(hyp * hyp + R * R) - aC
        return GeographicPosition(
            latitude=Angle(radians=lat),
            longitude=Angle(radians=lon),
            elevation=Distance(height_au),
            itrs_xyz=Distance(xyz_au),
            model=self,
        )

    def subpoint_of(self, position):
        """Return the point on the ellipsoid directly below a ``position``.

        The position’s ``.center`` must be 399, the center of the Earth.
        Returns a `GeographicPosition` giving the geodetic ``latitude``
        and ``longitude`` that lie directly below the input position,
        and an ``elevation`` above the ellipsoid of zero.

        """
        xyz_au, x, y, R, aC, hyp, lat = self._compute_latitude(position)
        lon = (arctan2(y, x) - pi) % tau - pi
        return self.latlon(lat * RAD2DEG, lon * RAD2DEG)

    def _compute_latitude(self, position):
        if position.center != 399:
            raise ValueError(
                'you can only calculate a geographic position from a'
                ' position which is geocentric (center=399), but this'
                ' position has a center of {0}'.format(position.center)
            )
        xyz_au = position.frame_xyz(itrs).au
        x, y, z = xyz_au
        a = self.radius.au
        e2 = self._e2
        R = sqrt(x*x + y*y)
        lat = arctan2(z, R)
        for iteration in 0,1,2:
            sin_lat = sin(lat)
            e2_sin_lat = e2 * sin_lat
            # At 0°, aC = 6378 km, Earth's actual radius at the equator.
            # At 90°, aC = 6399 km, Earth's radius of curvature at the pole.
            aC = a / sqrt(1.0 - e2_sin_lat * sin_lat)
            hyp = z + aC * e2_sin_lat
            lat = arctan2(hyp, R)
        return xyz_au, x, y, R, aC, hyp, lat

    subpoint = geographic_position_of  # deprecated method name

wgs84 = Geoid('WGS84', 6378137.0, 298.257223563)
iers2010 = Geoid('IERS2010', 6378136.6, 298.25642)

wgs84.__doc__ = """World Geodetic System 1984 `Geoid`.

This is the standard geoid used by the GPS system,
and is likely the standard that’s intended
if you are supplied a latitude and longitude
that don’t specify an alternative geoid.

"""
iers2010.__doc__ = 'International Earth Rotation Service 2010 `Geoid`.'

# Compatibility with old versions of Skyfield:

class Topos(GeographicPosition):
    """Deprecated: use ``wgs84.latlon()`` or ``iers2010.latlon()`` instead."""

    def __init__(self, latitude=None, longitude=None, latitude_degrees=None,
                 longitude_degrees=None, elevation_m=0.0, x=0.0, y=0.0):

        if latitude_degrees is not None:
            pass
        elif isinstance(latitude, Angle):
            latitude_degrees = latitude.degrees
        elif isinstance(latitude, (str, float, tuple)):
            latitude_degrees = _ltude(latitude, 'latitude', 'N', 'S')
        else:
            raise TypeError('please provide either latitude_degrees=<float>'
                            ' or latitude=<skyfield.units.Angle object>'
                            ' with north being positive')

        if longitude_degrees is not None:
            pass
        elif isinstance(longitude, Angle):
            longitude_degrees = longitude.degrees
        elif isinstance(longitude, (str, float, tuple)):
            longitude_degrees = _ltude(longitude, 'longitude', 'E', 'W')
        else:
            raise TypeError('please provide either longitude_degrees=<float>'
                            ' or longitude=<skyfield.units.Angle object>'
                            ' with east being positive')

        # Sneaky: the model thinks it's creating an object when really
        # it's just calling our superclass __init__() for us.  Alas, the
        # crimes committed to avoid duplicating code!  (This is actually
        # quite clean compared to the other alternatives I tried.)
        iers2010.latlon(latitude_degrees, longitude_degrees, elevation_m,
                        super(Topos, self).__init__)

        self.R_lat = self._R_lat  # On this old class, it was public.

    def itrf_xyz(self):
        return self.itrs_xyz



"""Routines solving basic geometric problems in astronomy."""

from .functions import dots, length_of, nan, sqrt_nan, where

def intersect_line_and_sphere(endpoint, center, radius):
    """Compute distance to intersections of a line and a sphere.

    Given a line through the origin (0,0,0) and an |xyz| ``endpoint``,
    and a sphere with the |xyz| ``center`` and scalar ``radius``,
    return the distance from the origin to their two intersections.

    If the line is tangent to the sphere, the two intersections will be
    at the same distance.  If the line does not intersect the sphere,
    two ``nan`` values will be returned.

    """
    # See http://paulbourke.net/geometry/circlesphere/index.html#linesphere
    # Names "b" and "c" designate the familiar values from the quadratic
    # formula; happily, a = 1 because we use a unit vector for the line.

    minus_b = 2.0 * (endpoint / length_of(endpoint) * center).sum(axis=0)
    c = (center * center).sum(axis=0) - radius * radius
    discriminant = minus_b * minus_b - 4 * c
    dsqrt = sqrt_nan(discriminant)
    return (minus_b - dsqrt) / 2.0, (minus_b + dsqrt) / 2.0

def line_and_ellipsoid_intersection(line_start, line_direction, radii):
    """Return the |xyz| position where a line intersects an ellipsoid.

    All three arguments are |xyz| arrays.  The line is specified by a
    ``line_start`` endpoint and a ``line_direction`` vector.  The
    ellipsoid is centered at the origin and is specified by its three
    ``radii`` that point along the three coordinate axes.

    Returns the |xyz| point of intersection, or ``[nan nan nan]`` if the
    line does not intersect the sphere.

    """
    # Based on `surfpt.f` from the SPICE Toolkit.
    if len(getattr(line_start, 'shape', ())) > 1:
        radii = radii.reshape((3, 1))

    # Scale coordinates so the ellipsoid becomes the unit sphere.
    start = line_start / radii
    direction = line_direction / radii

    # Where does the line come closest to the sphere's center?
    closest_point = start - _vector_projection(start, direction)

    startmag = length_of(start)
    pmag = length_of(closest_point)

    is_inside_sphere = startmag < 1.0
    is_behind_us = dots(closest_point - start, direction) < 0.0

    sign = where(is_inside_sphere, +1.0, where(is_behind_us, nan, -1.0))
    half_chord_length = sqrt_nan(1.0 - pmag*pmag)
    unit_direction = direction / length_of(direction)
    intersection = closest_point + sign * half_chord_length * unit_direction
    return where(startmag == 1.0, line_start, intersection * radii)

def _vector_projection(a, b):
    return dots(a,b) / dots(b,b) * b



"""Top-level objects and functions offered by the Skyfield library.

Importing this library is not always the fastest way to use a Skyfield
feature, since importing this module involves importing almost the
entirety of Skyfield and its dependencies, but is the most convenient
way for most users to use Skyfield's main features.

"""
from datetime import datetime
from .constants import B1950, T0, pi, tau
from .constellationlib import load_constellation_map, load_constellation_names
from .iokit import Loader, load_file
from .planetarylib import PlanetaryConstants
from .positionlib import SSB, position_from_radec, position_of_radec
from .starlib import Star
from .sgp4lib import EarthSatellite
from .timelib import (
    GREGORIAN_START, GREGORIAN_START_ENGLAND, Time, Timescale, utc
)
from .toposlib import Topos, iers2010, wgs84
from .units import Angle, Distance, Velocity, wms

load = Loader('.')
N = E = +1.0
S = W = -1.0

__all__ = [
    'Angle', 'B1950', 'Distance', 'E', 'EarthSatellite',
    'GREGORIAN_START', 'GREGORIAN_START_ENGLAND',
    'Loader', 'PlanetaryConstants', 'N', 'S', 'SSB', 'Star', 'W',
    'T0', 'Time', 'Timescale', 'Topos', 'Velocity',
    'datetime', 'iers2010', 'load', 'load_constellation_map',
    'load_constellation_names', 'load_file',
    'position_from_radec', 'position_of_radec',
    'utc', 'pi', 'tau', 'wgs84', 'wms',
]



# -*- coding: utf-8 -*-
"""Classes representing different kinds of astronomical position."""

from jplephem.names import target_names as target_names
from numpy import array, cos, einsum, full, reshape, nan, nan_to_num, zeros

from . import framelib
from .constants import ANGVEL, AU_M, C, C_AUDAY, ERAD, DAY_S, RAD2DEG, pi, tau
from .descriptorlib import reify
from .earthlib import compute_limb_angle
from .functions import (
    _T, _to_array, _to_spherical_and_rates, angle_between, from_spherical,
    length_of, mxm, mxv, rot_z, to_spherical,
)
from .geometry import intersect_line_and_sphere
from .relativity import (
    _compute_deflection, _compute_deflector_position, add_aberration, rmasses,
)
from .timelib import Time
from .units import Angle, AngleRate, Distance, Velocity, _interpret_angle

_GIGAPARSEC_AU = 206264806247096.38  # 1e9 * 360 * 3600 / tau

def build_position(position_au, velocity_au_per_d=None, t=None,
                   center=None, target=None):
    if center == 0:
        cls = Barycentric
    elif center == 399:
        cls = Geocentric
    else:
        cls = ICRF
    return cls(position_au, velocity_au_per_d, t, center, target)

def position_of_radec(ra_hours, dec_degrees, distance_au=_GIGAPARSEC_AU,
                      epoch=None, t=None, center=None, target=None):
    """Build a position object from a right ascension and declination.

    If a specific ``distance_au`` is not provided, Skyfield returns a
    position vector a gigaparsec in length.  This puts the position at a
    great enough distance that it will stand at the same right ascension
    and declination from any viewing position in the Solar System, to
    very high precision (within a few hundredths of a microarcsecond).

    If an ``epoch`` is specified, the input coordinates are understood
    to be in the dynamical system of that particular date.  Otherwise,
    they will be assumed to be ICRS (the modern replacement for J2000).

    .. versionadded:: 1.21
       This replaces a deprecated function ``position_from_radec()``
       whose ``distance`` argument was not as well designed.

    """
    theta = _to_array(dec_degrees) / 360.0 * tau
    phi = _to_array(ra_hours) / 24.0 * tau
    position_au = from_spherical(distance_au, theta, phi)
    if epoch is not None:
        position_au = mxv(epoch.MT, position_au)
    return build_position(position_au, None, t, center, target)

def position_from_radec(ra_hours, dec_degrees, distance=1.0, epoch=None,
                        t=None, center=None, target=None):
    """DEPRECATED version of ``position_of_radec()``."""
    return position_of_radec(ra_hours, dec_degrees, distance, epoch,
                             t, center, target)

class SSB(object):
    """The Solar System Barycenter."""
    @staticmethod
    def at(t):
        """Return the position of the Solar System Barycenter at time ``t``."""
        shape = (3,)
        if t.shape:
            shape += t.shape
        return Barycentric(zeros(shape), zeros(shape), t)

class ICRF(object):
    """An |xyz| position and velocity oriented to the ICRF axes.

    The International Celestial Reference Frame (ICRF) is a permanent
    reference frame that is the replacement for J2000.  Their axes agree
    to within 0.02 arcseconds.  It also supersedes older equinox-based
    systems like B1900 and B1950.

    Each instance of this class provides a ``.xyz`` vector and a
    ``.velocity`` vector that specify |xyz| coordinates along the axes
    of the ICRF.  A specific time ``.t`` might be specified or might be
    ``None``.

    """
    center_barycentric = None
    _observer_gcrs_au = None
    _default_center = None
    _ephemeris = None  # cached so we can compute how light is deflected

    def __init__(self, position_au, velocity_au_per_d=None, t=None,
                 center=None, target=None):
        self.t = t
        self.position = self.xyz = Distance(position_au)
        if velocity_au_per_d is None:
            velocity_au_per_d = full(self.xyz.au.shape, nan)
        self.velocity = Velocity(velocity_au_per_d)
        self.center = self._default_center if center is None else center
        self.target = target
        if center == 0:
            self.center_barycentric = self

    @classmethod
    def from_radec(cls, ra_hours, dec_degrees,
                   distance_au=_GIGAPARSEC_AU, epoch=None):
        theta = _to_array(dec_degrees) / 360.0 * tau
        phi = _to_array(ra_hours) / 24.0 * tau
        position_au = from_spherical(distance_au, theta, phi)
        if epoch is not None:
            position_au = mxv(epoch.MT, position_au)
        return cls(position_au)

    @classmethod
    def from_time_and_frame_vectors(cls, t, frame, distance, velocity):
        """Constructor: build a position from two vectors in a reference frame.

        * ``t`` — The :class:`~skyfield.timelib.Time` of the position.
        * ``frame`` — A reference frame listed at `reference_frames`.
        * ``distance`` — A `Distance` |xyz| vector in the given frame.
        * ``velocity`` — A `Velocity` ẋ,ẏ,ż vector in the given frame.

        """
        r, v = distance.au, velocity.au_per_d
        at = getattr(frame, '_dRdt_times_RT_at', None)
        if at is not None:
            V = at(t)
            v = v - mxv(V, r)  # subtract instead of transposing
        RT = _T(frame.rotation_at(t))
        r = mxv(RT, r)
        v = mxv(RT, v)
        return cls(r, v, t)  # TODO: args for center and target?

    def __repr__(self):
        name = self.__class__.__name__
        center = self.center
        if name == 'Barycentric' and center == 0:
            suffix = ' BCRS'
        elif name == 'Apparent' and center == 399:
            suffix = ' GCRS'
        elif name != 'ICRF':
            suffix = ' ICRS'
        else:
            suffix = ''

        center = self.center
        target = self.target

        center_name = getattr(center, 'target_name', None)
        if center_name is None:
            center_name = str(center)

        target_name = getattr(target, 'target_name', None)
        if target_name is None:
            target_name = str(target)

        return '<{0}{1} position{2}{3}{4}{5}>'.format(
            name,
            suffix,
            '' if (self.velocity is None) else ' and velocity',
            '' if self.t is None else ' at date t',
            '' if self.center is None else ' center={0}'.format(center_name),
            '' if self.target is None else ' target={0}'.format(target_name),
        )

    def __sub__(self, body):
        """Subtract two ICRF vectors to produce a third."""
        if self.center != body.center:
            raise ValueError(
                "you can only subtract two vectors"
                " if they both start at the same center"
            )
        p = self.xyz.au - body.xyz.au
        if self.velocity is None or body.velocity is None:
            v = None
        else:
            v = self.velocity.au_per_d - body.velocity.au_per_d
        return build_position(p, v, self.t, body.target, self.target)

    def __getitem__(self, i):
        return type(self)(
            self.xyz.au[:,i],
            self.velocity.au_per_d[:,i],
            self.t[i],
            self.center,
            self.target,
        )

    def __neg__(self):
        return type(self)(
            -self.xyz.au,
            -self.velocity.au_per_d,
            self.t,
            self.target,
            self.center,
        )

    def distance(self):
        """Compute the distance from the origin to this position.

        The return value is a :class:`~skyfield.units.Distance` that
        prints itself out in astronomical units (au) but that also
        offers attributes ``au``, ``km``, and ``m`` if you want to
        access its magnitude as a number.

        >>> v = ICRF([1, 1, 0])
        >>> print(v.distance())
        1.41421 au

        """
        return Distance(length_of(self.xyz.au))

    def speed(self):
        """Compute the magnitude of the velocity vector.

        >>> v = ICRF([0, 0, 0], [1, 2, 3])
        >>> print(v.speed())
        3.74166 au/day

        """
        return Velocity(length_of(self.velocity.au_per_d))

    @reify
    def light_time(self):
        """Length of this vector in days of light travel time."""
        # Note that this property is almost never called, since
        # .light_time is set manually on the one kind of position
        # (astrometric) that is ever likely to need it.  Alas: back in
        # 2015, I didn't think to either hide this attribute, or have it
        # include its units in its name!
        return self.distance().m / C / DAY_S

    def radec(self, epoch=None):
        r"""Compute equatorial RA, declination, and distance.

        When called without a parameter, this returns standard ICRF
        right ascension and declination:

        >>> from skyfield.api import load
        >>> ts = load.timescale()
        >>> t = ts.utc(2020, 5, 13, 10, 32)
        >>> eph = load('de421.bsp')
        >>> astrometric = eph['earth'].at(t).observe(eph['sun'])
        >>> ra, dec, distance = astrometric.radec()
        >>> print(ra, dec, sep='\n')
        03h 21m 47.67s
        +18deg 28' 55.3"

        If you instead want the coordinates referenced to the dynamical
        system defined by the Earth's true equator and equinox, provide
        a specific epoch time.

        >>> ra, dec, distance = astrometric.apparent().radec(epoch='date')
        >>> print(ra, dec, sep='\n')
        03h 22m 54.73s
        +18deg 33' 04.5"

        """
        position_au = self.xyz.au
        if epoch is not None:
            if isinstance(epoch, Time):
                pass
            elif isinstance(epoch, float):
                epoch = Time(None, tt=epoch)
            elif epoch == 'date':
                epoch = self.t
            else:
                raise ValueError('the epoch= must be a Time object,'
                                 ' a floating point Terrestrial Time (TT),'
                                 ' or the string "date" for epoch-of-date')
            position_au = mxv(epoch.M, position_au)
        r_au, dec, ra = to_spherical(position_au)
        return (Angle(radians=ra, preference='hours'),
                Angle(radians=dec, signed=True),
                Distance(r_au))

    def hadec(self):
        """Compute hour angle, declination, and distance.

        Returns a tuple of two :class:`~skyfield.units.Angle` objects
        plus the :class:`~skyfield.units.Distance` to the target.  The
        angles are the hour angle (±12 hours) east or west of your
        meridian along the ITRS celestial equator, and the declination
        (±90 degees) above or below it.  This only works for positions
        whose center is a geographic location; otherwise, there is no
        local meridian from which to measure the hour angle.

        Because this declination is measured from the plane of the
        Earth’s physical geographic equator, it will be slightly
        different than the declination returned by ``radec()`` if you
        have loaded a :ref:`polar-motion` file.

        The coordinates are not adjusted for atmospheric refraction near
        the horizon.

        """
        R = framelib.itrs.rotation_at(self.t)
        r = mxv(R, self.xyz.au)
        au, dec, sublongtiude = to_spherical(r)
        lon = getattr(self.center, 'longitude', None)
        if lon is None:
            raise ValueError(_hadec_message)
        ha = lon.radians - sublongtiude
        ha += pi
        ha %= tau
        ha -= pi
        return (Angle(radians=ha, preference='hours', signed=True),
                Angle(radians=dec, signed=True),
                Distance(au))

    def altaz(self, temperature_C=None, pressure_mbar='standard'):
        """Compute (alt, az, distance) relative to the observer's horizon

        The altitude returned is an :class:`~skyfield.units.Angle`
        measured in degrees above the horizon, while the azimuth
        :class:`~skyfield.units.Angle` measures east along the horizon
        from geographic north (so 0 degrees means north, 90 is east, 180
        is south, and 270 is west).

        By default, Skyfield does not adjust the altitude for
        atmospheric refraction.  If you want Skyfield to estimate how
        high the atmosphere might lift the body's image, give the
        argument ``temperature_C`` either the temperature in degrees
        centigrade, or the string ``'standard'`` (in which case 10°C is
        used).

        When calculating refraction, Skyfield uses the observer’s
        elevation above sea level to estimate the atmospheric pressure.
        If you want to override that value, simply provide a number
        through the ``pressure_mbar`` parameter.

        """
        return _to_altaz(self, temperature_C, pressure_mbar)

    def separation_from(self, another_icrf):
        """Return the angle between this position and another.

        >>> from skyfield.api import load
        >>> ts = load.timescale()
        >>> t = ts.utc(2020, 4, 18)
        >>> eph = load('de421.bsp')
        >>> sun, venus, earth = eph['sun'], eph['venus'], eph['earth']
        >>> e = earth.at(t)
        >>> s = e.observe(sun)
        >>> v = e.observe(venus)
        >>> print(s.separation_from(v))
        43deg 23' 23.1"

        You can also compute separations across an array of positions.

        >>> t = ts.utc(2020, 4, [18, 19, 20])
        >>> e = earth.at(t)
        >>> print(e.observe(sun).separation_from(e.observe(venus)))
        3 values from 43deg 23' 23.1" to 42deg 49' 46.6"

        """
        u = self.xyz.au
        v = another_icrf.xyz.au

        # Allow an array of positions to be compared with a single other
        # position.
        difference = len(u.shape) - len(v.shape)
        if difference:
            if difference > 0:
                v = reshape(v, v.shape + (1,) * difference)
            else:
                u = reshape(u, u.shape + (1,) * -difference)

        return Angle(radians=angle_between(u, v))

    # TODO: build a reference frame for the following two methods.

    def cirs_xyz(self, epoch):
        """Compute cartesian CIRS coordinates at a given epoch |xyz|.

        Calculate coordinates in the Celestial Intermediate Reference System
        (CIRS), a dynamical coordinate system referenced to the Celestial
        Intermediate Origin (CIO). As this is a dynamical system it must be
        calculated at a specific epoch.
        """
        if isinstance(epoch, Time):
            pass
        elif isinstance(epoch, float):
            epoch = Time(None, tt=epoch)
        elif epoch == 'date':
            epoch = self.t
        else:
            raise ValueError('the epoch= must be a Time object,'
                             ' a floating point Terrestrial Time (TT),'
                             ' or the string "date" for epoch-of-date')

        vector = mxv(epoch.C, self.xyz.au)
        return Distance(vector)

    def cirs_radec(self, epoch):
        """Get spherical CIRS coordinates at a given epoch (ra, dec, distance).

        Calculate coordinates in the Celestial Intermediate Reference System
        (CIRS), a dynamical coordinate system referenced to the Celestial
        Intermediate Origin (CIO). As this is a dynamical system it must be
        calculated at a specific epoch.
        """
        r_au, dec, ra = to_spherical(self.cirs_xyz(epoch).au)

        return (Angle(radians=ra, preference='hours'),
                Angle(radians=dec, signed=True),
                Distance(r_au))

    # Deprecated methods, that have been replaced by `framelib.py` plus
    # the "frame" methods in the next section.

    def ecliptic_xyz(self, epoch=None):
        if epoch is None:
            return self.frame_xyz(framelib.ecliptic_J2000_frame)
        return _Fake(self, epoch).frame_xyz(framelib.ecliptic_frame)
    def ecliptic_velocity(self):
        return self.frame_xyz_and_velocity(framelib.ecliptic_J2000_frame)[1]
    def ecliptic_latlon(self, epoch=None):
        if epoch is None:
            return self.frame_latlon(framelib.ecliptic_J2000_frame)
        return _Fake(self, epoch).frame_latlon(framelib.ecliptic_frame)

    def galactic_xyz(self): return self.frame_xyz(framelib.galactic_frame)
    def galactic_latlon(self): return self.frame_latlon(framelib.galactic_frame)
    ecliptic_position = ecliptic_xyz  # old alias
    galactic_position = galactic_xyz  # old alias

    # New methods for converting to and from `framelib.py` reference frames.

    def frame_xyz(self, frame):
        """Return this position as an |xyz| vector in a reference frame.

        Returns a :class:`~skyfield.units.Distance` object giving the
        |xyz| of this position in the given ``frame``.  See
        `reference_frames`.

        """
        return Distance(mxv(frame.rotation_at(self.t), self.xyz.au))

    def frame_xyz_and_velocity(self, frame):
        """Return |xyz| position and velocity vectors in a reference frame.

        Returns two vectors in the given coordinate ``frame``: a
        :class:`~skyfield.units.Distance` providing an |xyz| position
        and a :class:`~skyfield.units.Velocity` giving (xdot,ydot,zdot)
        velocity.  See `reference_frames`.

        """
        R = frame.rotation_at(self.t)
        r, v = self.xyz.au, self.velocity.au_per_d
        r = mxv(R, r)
        v = mxv(R, v)
        at = getattr(frame, '_dRdt_times_RT_at', None)
        if at is not None:
            V = at(self.t)
            v += mxv(V, r)
        return Distance(r), Velocity(v)

    def frame_latlon(self, frame):
        """Return latitude, longitude, and distance in the given frame.

        Returns a 3-element tuple giving the latitude and longitude as
        :class:`~skyfield.units.Angle` objects and the range to the
        target as a :class:`~skyfield.units.Distance`.  See
        `reference_frames`.

        """
        vector = mxv(frame.rotation_at(self.t), self.xyz.au)
        d, lat, lon = to_spherical(vector)
        return (Angle(radians=lat, signed=True),
                Angle(radians=lon),
                Distance(d))

    def frame_latlon_and_rates(self, frame):
        """Return a reference frame latitude, longitude, range, and rates.

        Return a 6-element tuple of 3 coordinates and 3 rates-of-change
        for this position in the given reference ``frame``:

        * Latitude :class:`~skyfield.units.Angle` from +90° north to −90° south
        * Longitude :class:`~skyfield.units.Angle` 0°–360° east
        * Radial :class:`~skyfield.units.Distance`
        * Latitude :class:`~skyfield.units.AngleRate`
        * Longitude :class:`~skyfield.units.AngleRate`
        * Radial :class:`~skyfield.units.Velocity`

        If the reference frame is the ICRS, or is J2000, or otherwise
        involves the celestial equator and pole, then the latitude and
        longitude returned will measure what are more commonly called
        “declination” and “right ascension”.  Note that right ascension
        is usually expressed as hours (24 in a circle), rather than in
        the degrees that this routine will return.

        """
        r, v = self.frame_xyz_and_velocity(frame)
        r = r.au
        v = v.au_per_d
        d, lat, lon, d_rate, lat_rate, lon_rate = _to_spherical_and_rates(r, v)
        return (Angle(radians=lat, signed=True),
                Angle(radians=lon),
                Distance(d),
                AngleRate._from_radians_per_day(lat_rate),
                AngleRate._from_radians_per_day(lon_rate),
                Velocity(d_rate))

    def to_skycoord(self, unit=None):
        """Convert this distance to an AstroPy ``SkyCoord`` object.

        Currently, this will only work with Skyfield positions whose
        center is the Solar System barycenter or else the geocenter.

        """
        from astropy.coordinates import SkyCoord
        from astropy.units import au
        if self.center == 0:
            frame = 'icrs'
        elif self.center == 399:
            from astropy.coordinates import GCRS
            frame = GCRS(obstime=self.t.to_astropy())
        else:
            raise NotImplementedError(
                'Skyfield can only return AstroPy positions for barycentric'
                ' and geocentric coordinates; if you are interested in adding'
                ' support for positions whose center={}, please open an issue'
                .format(self.center)
            )

        x, y, z = self.xyz.au
        return SkyCoord(frame=frame, representation_type='cartesian',
                        unit=au, x=x, y=y, z=z)

    def phase_angle(self, sun):
        """Return this position’s phase angle: the angle Sun-target-observer.

        Given a Sun object (which you can build by loading an ephemeris
        and looking up ``eph['Sun']``), return the `Angle` from the
        body's point of view between light arriving from the Sun and the
        light departing toward the observer.  This angle is 0° if the
        observer is in the same direction as the Sun and sees the body
        as fully illuminated, and 180° if the observer is behind the
        body and sees only its dark side.

        .. versionadded:: 1.42

        """
        # TODO: should we have the body .observe() the Sun, to account
        # for light travel time?
        s = sun.at(self.t)
        u = self.xyz.au
        v = u - s.xyz.au + self.center_barycentric.xyz.au
        return Angle(radians=angle_between(u, v))

    def fraction_illuminated(self, sun):
        """Return the fraction of the target’s disc that is illuminated.

        Given a Sun object (which you can build by loading an ephemeris
        and looking up ``eph['Sun']``), compute what fraction from 0.0
        to 1.0 of this target’s disc is illuminated, under the
        assumption that the target is a sphere.

        .. versionadded:: 1.42

        """
        a = self.phase_angle(sun).radians
        return 0.5 * (1.0 + cos(a))

    def is_sunlit(self, ephemeris):
        """Return whether a position in Earth orbit is in sunlight.

        Returns ``True`` or ``False``, or an array of such values, to
        indicate whether this position is in sunlight or is blocked by
        the Earth’s shadow.  It should work with positions produced
        either by calling ``at()`` on a satellite object, or by calling
        ``at()`` on the relative position ``sat - topos`` of a satellite
        with respect to an Earth observer’s position.  See
        :ref:`satellite-is-sunlit`.

        """
        if self.center == 399:
            earth_m = - self.xyz.m
        else:
            gcrs_position = self._observer_gcrs_au
            if gcrs_position is None:
                raise ValueError('cannot tell whether this position is sunlit')
            earth_m = - self.xyz.m - gcrs_position * AU_M

        sun_m = (ephemeris['sun'] - ephemeris['earth']).at(self.t).xyz.m
        near, far = intersect_line_and_sphere(sun_m + earth_m, earth_m, ERAD)
        return nan_to_num(far) <= 0

    def is_behind_earth(self):
        """Return whether the Earth blocks the view of this object.

        For a position centered on an Earth-orbiting satellite, return
        whether the target is in eclipse behind the disc of the Earth.
        See :ref:`is-behind-earth`.

        """
        observer_gcrs_au = self._observer_gcrs_au
        if observer_gcrs_au is None:
            raise ValueError('can only compute Earth occultation for'
                             ' positions observed from an Earth satellite')
        earth_m = - observer_gcrs_au * AU_M
        vector_m = self.xyz.m
        near, far = intersect_line_and_sphere(vector_m, earth_m, ERAD)
        length_m = length_of(vector_m)
        return (nan_to_num(far) > 0) & (nan_to_num(near) < length_m)

    @reify
    def _altaz_rotation(self):
        # Return and cache (with @reify) the orientation of this
        # observer, in case a single observer.at() position is used in
        # several subsequent .observe().apparent().altaz() calls.
        rotation_at = getattr(self.target, 'rotation_at', None)
        if rotation_at is None:
            raise ValueError(_altaz_message)
        return rotation_at(self.t)

    def from_altaz(self, alt=None, az=None, alt_degrees=None, az_degrees=None,
                   distance=Distance(au=0.1)):
        """Generate an Apparent position from an altitude and azimuth.

        The altitude and azimuth can each be provided as an `Angle`
        object, or else as a number of degrees provided as either a
        float or a tuple of degrees, arcminutes, and arcseconds::

            alt=Angle(...), az=Angle(...)
            alt_degrees=23.2289, az_degrees=142.1161
            alt_degrees=(23, 13, 44.1), az_degrees=(142, 6, 58.1)

        The distance should be a :class:`~skyfield.units.Distance`
        object, if provided; otherwise a default of 0.1 au is used.

        """
        # TODO: should this method live on another class?

        rotation_at = getattr(self.target, 'rotation_at', None)
        if rotation_at is None:
            raise ValueError(_altaz_message)
        R = rotation_at(self.t)

        alt = _interpret_angle('alt', alt, alt_degrees)
        az = _interpret_angle('az', az, az_degrees)
        r = distance.au
        p = from_spherical(r, alt, az)
        p = einsum('ji...,j...->i...', R, p)
        return Apparent(p)

# For compatibility with my original name for the class.  Not an
# important enough change to warrant a deprecation error for users, so:
ICRS = ICRF

class Geometric(ICRF): pass  # deprecated; kept for backwards compatibility

class Barycentric(ICRF):
    """An |xyz| position measured from the Solar System barycenter.

    Skyfield generates a `Barycentric` position measured from the
    gravitational center of the Solar System whenever you ask a body for
    its location at a particular time:

    >>> t = ts.utc(2003, 8, 29)
    >>> mars.at(t)
    <Barycentric BCRS position and velocity at date t center=0 target=499>

    This class’s ``.xyz`` and ``.velocity`` are |xyz| vectors in
    the Barycentric Celestial Reference System (BCRS), the modern
    replacement for J2000 coordinates measured from the Solar System
    Barycenter.

    """
    _default_center = 0

    def observe(self, body):
        """Compute the `Astrometric` position of a body from this location.

        To compute the body's astrometric position, it is first asked
        for its position at the time `t` of this position itself.  The
        distance to the body is then divided by the speed of light to
        find how long it takes its light to arrive.  Finally, the light
        travel time is subtracted from `t` and the body is asked for a
        series of increasingly exact positions to learn where it was
        when it emitted the light that is now reaching this position.

        >>> earth.at(t).observe(mars)
        <Astrometric ICRS position and velocity at date t center=399 target=499>

        """
        p, v, t, light_time = body._observe_from_bcrs(self)
        astrometric = Astrometric(p, v, t, self.target, body.target)
        astrometric._ephemeris = self._ephemeris
        astrometric.center_barycentric = self
        astrometric.light_time = light_time
        return astrometric

# TODO: pre-create a Barycentric object representing the SSB, and make
# it possible for it to observe() a planet.

class Astrometric(ICRF):
    """An astrometric |xyz| position relative to a particular observer.

    The astrometric position of a body is its position relative to an
    observer, adjusted for light-time delay.  It is the position of the
    body back when it emitted (or reflected) the light that is now
    reaching the observer's eye or telescope.  Astrometric positions are
    usually generated in Skyfield by calling the `Barycentric` method
    `observe()`, which performs the light-time correction.

    Both the ``.xyz`` and ``.velocity`` are |xyz| vectors
    oriented along the axes of the ICRF, the modern replacement for the
    J2000 reference frame.

    It is common to either call ``.radec()`` (with no argument) on an
    astrometric position to generate an *astrometric place* right
    ascension and declination with respect to the ICRF axes, or else to
    call ``.apparent()`` to generate an :class:`Apparent` position.

    """
    def altaz(self):
        raise ValueError(
            'it is not useful to call .altaz() on an astrometric position;'
            ' try calling .apparent() first to get an apparent position'
        )

    def apparent(self, deflectors=(10, 599, 699)):
        """Compute an :class:`Apparent` position for this body.

        This applies two effects to an astrometric position that arise
        from relativity and that shift the position slightly: deflection
        as light passes massive bodies on its way to the observer, and
        the aberration of light caused by the observer's own velocity.

        These two transforms convert the position from the BCRS
        reference frame of the Solar System barycenter to the reference
        frame of the observer.  In the specific case where the observer
        is the Earth, the output reference frame is the GCRS.

        The default value of the ``deflectors`` argument selects the
        three largest Solar System masses for deflection: the Sun 10,
        Jupiter 599, and Saturn 699.  The caller can override this by
        providing their own tuple of codes, or an empty tuple.

        If the observer is on the Earth's surface or in Earth orbit,
        then Skyfield also applies the deflection caused by the Earth's
        own mass.

        """
        t = self.t
        target_au = self.xyz.au.copy()

        cb = self.center_barycentric
        bcrs_position = cb.xyz.au
        bcrs_velocity = cb.velocity.au_per_d
        observer_gcrs_au = cb._observer_gcrs_au

        # If a single observer position (3,) is observing an array of
        # targets (3,n), then deflection and aberration will complain
        # that "operands could not be broadcast together" unless we give
        # the observer another dimension too.
        if len(bcrs_position.shape) < len(target_au.shape):
            shape = bcrs_position.shape + (1,)
            bcrs_position = bcrs_position.reshape(shape)
            bcrs_velocity = bcrs_velocity.reshape(shape)
            if observer_gcrs_au is not None:
                observer_gcrs_au = observer_gcrs_au.reshape(shape)

        tlt = length_of(target_au) / C_AUDAY

        for code in deflectors:  # TODO: skip ourselves?
            rmass = rmasses[code]
            if (code % 100 == 99) and (code not in self._ephemeris):
                code //= 100
            try:
                deflector = self._ephemeris[code]
            except KeyError:
                m = _deflector_message.format(code, target_names[code])
                e = ValueError(m)
                e.__cause__ = None
                e.target = code
                raise e
            deflector_au = _compute_deflector_position(
                t, bcrs_position, target_au, deflector, tlt,
            )
            d = _compute_deflection(target_au, deflector_au, rmass)
            target_au += d

        if observer_gcrs_au is not None:
            rmass = rmasses[399]
            d = _compute_deflection(target_au, observer_gcrs_au, rmass)

            _, nadir_angle = compute_limb_angle(target_au, observer_gcrs_au)
            target_au += d * (nadir_angle >= 0.8)

        add_aberration(target_au, bcrs_velocity, self.light_time)

        # Note that aberration and deflection are not applied to velocity.
        v = self.velocity.au_per_d

        apparent = Apparent(target_au, v, t, self.center, self.target)
        apparent.center_barycentric = self.center_barycentric
        apparent._observer_gcrs_au = observer_gcrs_au
        return apparent

_deflector_message = """\
ephemeris is missing '{0} {1}'

By default, apparent() computes how the target's light is deflected by
the gravity of the Sun, Jupiter, and Saturn.  Either add the missing
body to your ephemeris, or adjust your call to apparent() like this:

    position.apparent(deflectors=(10, 599))  # only Sun and Jupiter
    position.apparent(deflectors=())         # turn deflection off

"""

class Apparent(ICRF):
    """An apparent |xyz| position relative to a particular observer.

    This class’s vectors provide the position and velocity of a body
    relative to an observer, adjusted to predict where the body’s image
    will really appear (hence "apparent") in the sky:

    * Light-time delay, as already present in an `Astrometric` position.

    * Deflection: gravity bends light, and thus the image of a distant
      object, as the light passes massive objects like Jupiter, Saturn,
      and the Sun.  For an observer on the Earth’s surface or in Earth
      orbit, the slight deflection by the gravity of the Earth itself is
      also included.

    * Aberration: incoming light arrives slanted because of the
      observer's motion through space.

    These positions are usually produced in Skyfield by calling the
    `apparent()` method of an `Astrometric` object.

    Both the ``.xyz`` and ``.velocity`` are |xyz| vectors
    oriented along the axes of the ICRF, the modern replacement for the
    J2000 reference frame.  If the observer is at the geocenter, they
    are more specifically GCRS coordinates.  Two common coordinates that
    this vector can generate are:

    * *Proper place:* call ``.radec()`` without arguments to compute
      right ascension and declination with respect to the fixed axes of
      the ICRF.

    * *Apparent place,* the most popular option: call ``.radec('date')``
      to generate right ascension and declination with respect to the
      equator and equinox of date.

    """

class Geocentric(ICRF):
    """An |xyz| position measured from the center of the Earth.

    A geocentric position is the difference between the position of the
    Earth at a given instant and the position of a target body at the
    same instant, without accounting for light-travel time or the effect
    of relativity on the light itself.

    Its ``.xyz`` and ``.velocity`` vectors have |xyz| axes that
    are those of the Geocentric Celestial Reference System (GCRS), an
    inertial system that is an update to J2000 and that does not rotate
    with the Earth itself.

    """
    _default_center = 399

    def itrf_xyz(self):
        """Deprecated; instead, call ``.frame_xyz(itrs)``. \
        See `reference_frames`."""
        return self.frame_xyz(framelib.itrs)

    def subpoint(self):
        """Deprecated; instead, call either ``iers2010.subpoint(pos)`` or \
        ``wgs84.subpoint(pos)``."""
        from .toposlib import iers2010
        return iers2010.subpoint(self)

def _to_altaz(position, temperature_C, pressure_mbar):
    """Compute (alt, az, distance) relative to the observer's horizon."""
    cb = position.center_barycentric
    if cb is not None:
        R = cb._altaz_rotation
    else:
        rotation_at = getattr(position.center, 'rotation_at')
        if rotation_at is not None:
            R = rotation_at(position.t)
        else:
            raise ValueError(_altaz_message)

    position_au = mxv(R, position.xyz.au)
    r_au, alt, az = to_spherical(position_au)

    if temperature_C is None:
        alt = Angle(radians=alt)
    else:
        refract = getattr(position.center, 'refract', None)
        if refract is None:
            raise ValueError(_altaz_message)
        alt = position.center.refract(
            alt * RAD2DEG, temperature_C, pressure_mbar,
        )

    return alt, Angle(radians=az), Distance(r_au)

_hadec_message = (
    'to compute a body’s hour angle, you must observe it'
    ' from a specific latitude and longitude on Earth'
)
_altaz_message = (
    'to compute altitude and azimuth, you must observe it from a specific'
    ' latitude and longitude on Earth, or else from a location on another'
    ' Solar System body that you have loaded from a set of planetary constants'
)

class _Fake(ICRF):  # support for deprecated frame rotation methods above
    def __init__(self, original, epoch):
        self.xyz = original.xyz
        if isinstance(epoch, Time):
            self.t = epoch
        elif isinstance(epoch, float):
            self.t = Time(None, tt=epoch)
        elif epoch == 'date':
            self.t = original.t
        else:
            raise ValueError('the epoch= must be a Time object,'
                             ' a floating point Terrestrial Time (TT),'
                             ' or the string "date" for epoch-of-date')

def ITRF_to_GCRS(t, rITRF):  # Deprecated; for compatibility with old versions.
    return mxv(_T(framelib.itrs.rotation_at(t)), rITRF)

def ITRF_to_GCRS2(t, rITRF, vITRF, _high_accuracy=False):
    position = array(rITRF)
    velocity = array(vITRF)

    # TODO: This is expensive, and should be extensively trimmed to only
    # include the most important terms underlying GAST.  But it improves
    # the precision by something like 1e5 times when compared to using
    # the round number skyfield.constants.ANGVEL!
    #
    # See the test `test_velocity_in_ITRF_to_GCRS2()`.
    #
    if _high_accuracy:
        _one_second = 1.0 / DAY_S
        t_later = t.ts.tt_jd(t.whole, t.tt_fraction + _one_second)
        angvel = (t_later.gast - t.gast) / 24.0 * tau
    else:
        angvel = ANGVEL

    spin = rot_z(t.gast / 24.0 * tau)
    R = mxm(t.MT, spin)

    z = 0.0 * angvel

    V = array((
        (z,-DAY_S * angvel,z),
        (DAY_S * angvel,z,z),
        (z,z,z),
    ))

    velocity = velocity + mxv(V, position)
    position = mxv(R, position)
    velocity = mxv(R, velocity)

    return position, velocity



from skyfield.api import (
    load_constellation_map, load_constellation_names, position_of_radec,
)
from skyfield.functions import load_bundled_npy
from skyfield.timelib import Time, julian_date_of_besselian_epoch

def test_constellations():
    lookup = load_constellation_map()

    assert lookup(position_of_radec(0, 0)) == 'Psc'
    assert lookup(position_of_radec(360, 90)) == 'UMi'

    # (4.65, 0) used to be in Orion, but, precession
    assert lookup(position_of_radec(4.65, 0)) == 'Eri'
    assert lookup(position_of_radec(4.75, 0.3)) == 'Ori'

    # exploit extend() bug, issue 547
    B1875 = Time(None, julian_date_of_besselian_epoch(1875))
    assert lookup(position_of_radec(18.1747, 39., epoch=B1875)) == 'Her'
    assert lookup(position_of_radec(18.1753, 39., epoch=B1875)) == 'Lyr'

    # Verify we are not using the paper's original table, which had an error.
    star = position_of_radec(16.323165, -18.848203, epoch=B1875)
    assert lookup(star) == 'Sco'

    # Test vectorization.
    assert list(
        lookup(position_of_radec([4.65, 4.75], [0, 0.3]))
    ) == ['Eri', 'Ori']

def test_that_constellation_abbreviations_match():
    # This test should ideally not have to load the underlying data
    # source that is behind load_constellation_map(), but for the moment
    # I do not see a quick alternative.
    arrays = load_bundled_npy('constellations.npz')
    abbrevs1 = set(arrays['indexed_abbreviations'])

    abbrevs2 = {abbrev for abbrev, name in load_constellation_names()}
    assert abbrevs1 == abbrevs2



"""Elegant astronomy for Python

Most users will use Skyfield by importing ``skyfield.api`` and using the
functions and classes there.  Visit https://rhodesmill.org/skyfield/ or
read the same content in the ``documentation`` directory in the repo.

"""
VERSION = 1, 54
__version__ = '.'.join(map(str, VERSION))



"""Test that Skyfield implements NOVAS-compatible Earth deflection.

Run `design/measure_earth_deflection.py` to draw a quick graph.

"""
from contextlib import contextmanager
from numpy import arange, diff
from skyfield import relativity
from skyfield.api import load, wgs84
from skyfield.framelib import ecliptic_frame as ef
from skyfield.jpllib import _jpl_name_code_dict as jpl_codes

@contextmanager
def patch_rmass(name):
    code = jpl_codes[name.upper()]
    old = relativity.rmasses[name]
    relativity.rmasses[code] = 1e100  # almost no mass
    relativity.rmasses[name] = 1e100
    try:
        yield
    finally:
        relativity.rmasses[code] = old
        relativity.rmasses[name] = old

def trial(deflector_name, target_name, year, month, day):
    """Compare apparent() lat/lon with a given deflector turned on and off."""
    ts = load.timescale()
    t = ts.tt(year, month, day, arange(0, 25, 6))
    eph = load('de421.bsp')
    earth = eph['earth']
    target = eph[target_name]
    lowell = earth + wgs84.latlon(35.2029, -111.6646)
    astrometric = lowell.at(t).observe(target)
    lat1, lon1, _ = astrometric.apparent().frame_latlon(ef)

    with patch_rmass(deflector_name):
        lat2, lon2, _ = astrometric.apparent().frame_latlon(ef)

    return lat1.mas() - lat2.mas(), lon1.mas() - lon2.mas()

def stringify(difference_mas):
    return ' '.join('%.2f' % n for n in difference_mas)

# The expected offsets for deflection in the following tests have NOT
# been confirmed against an external authority; they are simply what
# Skyfield 1.51 did when these tests were written.  But at least they
# stop us from accidentally changing Skyfield's behavior going forward.

def test_sun_deflection():
    lat_mas, lon_mas = trial('sun', 'jupiter barycenter', 2025, 3, 16)
    assert stringify(lat_mas) == '-0.01 -0.01 -0.01 -0.01 -0.01'
    assert stringify(lon_mas) == '4.16 4.17 4.19 4.21 4.22'

def test_jupiter_deflection():
    # Very close conjunction of 0.1 degrees.
    lat_mas, lon_mas = trial('jupiter', 'saturn barycenter', 2020, 12, 21)
    assert stringify(lat_mas) == '0.20 0.25 0.31 0.33 0.31'
    assert stringify(lon_mas) == '0.16 0.14 0.09 0.01 -0.08'

def test_saturn_deflection():
    # Conjunction, fairly distant 0.83 degrees separation.
    lat_mas, lon_mas = trial('saturn', 'neptune barycenter', 2026, 2, 20)
    assert stringify(lat_mas) == '0.01 0.01 0.01 0.01 0.01'
    assert stringify(lon_mas) == '0.00 0.00 0.00 -0.00 -0.00'

def test_moon_deflection():
    # Moon, the next deflector, should be turned off by default.
    lat_mas, lon_mas = trial('moon', 'neptune barycenter', 2024, 12, 9.364)
    assert stringify(lat_mas) == '0.00 0.00 0.00 0.00 0.00'
    assert stringify(lon_mas) == '0.00 0.00 0.00 0.00 0.00'

# Not tested against an external authority; but it's encouraging that
# Earth deflection affects altitude not azimuth, and bumps altitude up
# not down.

def test_earth_deflection():
    ts = load.timescale()
    t = ts.tt(2015, 3, 16, arange(0, 25, 2))
    eph = load('de421.bsp')
    earth = eph['earth']
    jupiter = eph['jupiter barycenter']
    lowell = earth + wgs84.latlon(35.2029, -111.6646)
    alt1, az1, _ = lowell.at(t).observe(jupiter).apparent().altaz()

    with patch_rmass('earth'):
        alt2, az2, _ = lowell.at(t).observe(jupiter).apparent().altaz()

    dalt = ' '.join('%.2f' % n for n in alt1.mas() - alt2.mas())
    daz = ' '.join('%.2f' % n for n in az1.mas() - az2.mas())

    s = '0.19 0.11 0.06 0.06 0.12 0.19 0.30 0.00 0.00 0.00 0.00 0.29 0.19'
    assert dalt == s

    s = '0.00 0.00 0.00 -0.00 -0.00 -0.00 -0.00 0.00 0.00 0.00 0.00 0.00 0.00'
    assert daz == s

def test_earth_deflection_cutoff():
    # The NOVAS library includes the Earth's gravitational deflection of
    # light for both topocentric observers and observers in Earth orbit,
    # but shuts this effect off once the object is behind the Earth 20%
    # of the way from its limb towards its center.  This test determines
    # whether Skyfield puts the resulting discontinuity in the same
    # place as the NOVAS library does.

    ts = load.timescale(delta_t=0.0)
    t = ts.tt(2016, 7, 2, arange(10.5628, 10.5639, 0.0002))
    planets = load('de405.bsp')
    earth = planets['earth']
    mars = planets['mars']
    lowell = earth + wgs84.latlon(35.2029, -111.6646)
    ra, dec, distance = lowell.at(t).observe(mars).apparent().radec()
    h = ra.hours
    hprime = diff(h)
    assert hprime[0] > 1.8e-8
    assert hprime[1] > 1.8e-8
    assert hprime[2] < 1.3e-8   # moment when nadir angle crosses 0.8
    assert hprime[3] > 1.8e-8
    assert hprime[4] > 1.8e-8



from numpy import array, isnan
from skyfield.geometry import intersect_line_and_sphere

def a(*args):
    return array(args)

def test_intersect_line_and_sphere():
    near, far = intersect_line_and_sphere(a(1000, 0, 0), a(2, 0, 0), 1)
    assert near == 1.0
    assert far == 3.0

    near, far = intersect_line_and_sphere(a(-1000, 0, 0), a(3, 0, 0), 1)
    assert near == -4.0
    assert far == -2.0

    near, far = intersect_line_and_sphere(a(0, 0.5, 0), a(0, 5, 0), 2)
    assert near == 3.0
    assert far == 7.0

    near, far = intersect_line_and_sphere(a(1000, 0, 0), a(0, 5, 0), 2)
    assert isnan(near)
    assert isnan(far)

    near, far = intersect_line_and_sphere(a(1000, 0, 0), a(2, 1, 0), 1)
    assert near == 2.0
    assert far == 2.0

    near, far = intersect_line_and_sphere(a(1000, 0, 0), a(2, 0.999, 0), 1)
    assert abs(near - 1.955) < 1e-3
    assert abs(far - 2.045) < 1e-3



"""Exceptions specific to the Skyfield library."""

class DeprecationError(Exception):
    """Explain that a Skyfield feature has been removed."""

class EphemerisRangeError(ValueError):
    """An ephemeris has been asked about positions outside its time range.

    Attributes:

    - `start_time`, `end_time`: the range of times supported by the segment
    - `time_mask`: Boolean array where ``True`` marks out-of-range times
    - `segment`: the ephemeris segment that was asked for positions

    """
    def __init__(self, message, start_time, end_time, time_mask, segment):
        self.args = message,
        self.start_time = start_time
        self.end_time = end_time
        self.time_mask = time_mask
        self.segment = segment



# -*- coding: utf-8 -*-
"""Raw transforms between coordinate frames, as NumPy matrices."""

from numpy import array
from .constants import ANGVEL, ASEC2RAD, DAY_S, tau
from .data.spice import inertial_frames as _inertial_frames
from .functions import mxm, rot_x, rot_z

def build_matrix():
    # 'xi0', 'eta0', and 'da0' are ICRS frame biases in arcseconds taken
    # from IERS (2003) Conventions, Chapter 5.

    xi0  = -0.0166170 * ASEC2RAD
    eta0 = -0.0068192 * ASEC2RAD
    da0  = -0.01460   * ASEC2RAD

    # Compute elements of rotation matrix.

    yx = -da0
    zx =  xi0
    xy =  da0
    zy =  eta0
    xz = -xi0
    yz = -eta0

    # Include second-order corrections to diagonal elements.

    xx = 1.0 - 0.5 * (yx * yx + zx * zx)
    yy = 1.0 - 0.5 * (yx * yx + zy * zy)
    zz = 1.0 - 0.5 * (zy * zy + zx * zx)

    return array(((xx, xy, xz), (yx, yy, yz), (zx, zy, zz)))

ICRS_to_J2000 = build_matrix()
del build_matrix

_identity = array([(1,0,0), (0,1,0), (0,0,1)])

class ICRS(object):
    """The International Celestial Reference System (ICRS).

    The ICRS is a permanent reference frame which has replaced J2000,
    with which its axes agree to within 0.02 arcseconds (closer than the
    precision of J2000 itself).  The ICRS also supersedes older
    equinox-based systems like B1900 and B1950.

    """
    @staticmethod
    def rotation_at(t):
        return _identity

class mean_equator_and_equinox_of_date(object):
    """The coordinate frame of Earth’s mean equator and equinox.

    This frame is used for measuring right ascension and declination.
    It tracks the Earth’s ‘mean’ equator and equinox which shift slowly
    across the sky due to precession, but ignores the smaller effects of
    nutation.

    """
    @staticmethod
    def rotation_at(t):
        return mxm(t.P, ICRS_to_J2000)

class true_equator_and_equinox_of_date(object):
    """The dynamical frame of Earth’s true equator and true equinox of date.

    This frame is used for measuring right ascension and declination.
    Unlike the fixed reference frames J2000 and the ICRS, this ‘TETE’
    frame rotates slowly as the Earth’s precession and nutation shift
    the equinox point.  Unlike the :class:`~skyfield.sgp4lib.TEME`
    frame, this frame doesn’t ignore nutation.

    This is supplied as an explicit reference frame in case you want
    |xyz| coordinates; if you want angles, it’s better to use the
    standard position method ``radec(epoch='date')`` since that will
    return the conventional units of hours-of-right-ascension instead of
    the degrees-of-longitude that ``frame_latlon()`` would return.

    This reference frame combines current theories of the Earth’s
    precession and nutation with a small offset between the ITRS and
    J2000 systems to produce right ascension and declination for a given
    date relative to the Earth’s axis and equator of rotation.

    """
    @staticmethod
    def rotation_at(t):
        return t.M

_itrs_angvel_matrix = array((
    (0.0, DAY_S * ANGVEL, 0.0),
    (-DAY_S * ANGVEL, 0.0, 0.0),
    (0.0, 0.0, 0.0),
))

class tirs(object):
    """The Terrestrial Intermediate Reference System (TIRS).

    Coordinates in this Earth-centered Earth-fixed (ECEF) system are
    measured from the axis and equator of the Earth’s rotation, ignoring
    the few tenths of an arcsecond by which the Earth’s actual crust and
    continents might be askance from the axis.  (More precisely: like
    the ITRS this frame accounts for precession and nutation, but
    neglects polar motion and the TIO locator.)

    """
    @staticmethod
    def rotation_at(t):
        return mxm(rot_z(-t.gast * tau / 24.0), t.M)

    @staticmethod
    def _dRdt_times_RT_at(t):
        # TODO: taking the derivative of the instantaneous angular
        # velocity provides a more accurate transform.
        return _itrs_angvel_matrix

tirs = tirs()

class itrs(object):
    """The International Terrestrial Reference System (ITRS).

    This is the IAU standard for an Earth-centered Earth-fixed (ECEF)
    coordinate system, anchored to the Earth’s crust and continents.
    This reference frame combines three other reference frames: the
    Earth’s true equator and equinox of date, the Earth’s rotation with
    respect to the stars, and (if your ``Timescale`` has polar offsets
    loaded) the polar wobble of the crust with respect to the Earth’s
    pole of rotation.

    .. versionadded:: 1.34

    """
    @staticmethod
    def rotation_at(t):
        R = mxm(rot_z(-t.gast * tau / 24.0), t.M)
        if t.ts.polar_motion_table is not None:
            R = mxm(t.polar_motion_matrix(), R)
        return R

    @staticmethod
    def _dRdt_times_RT_at(t):
        # TODO: taking the derivative of the instantaneous angular
        # velocity provides a more accurate transform.
        return _itrs_angvel_matrix

itrs = itrs()

def build_ecliptic_matrix(t):
    # Build the matrix to rotate an ICRF vector into ecliptic coordinates.
    _, d_eps = t._nutation_angles_radians
    true_obliquity = t._mean_obliquity_radians + d_eps
    return mxm(rot_x(- true_obliquity), t.M)

class ecliptic_frame(object):
    """Reference frame of the true ecliptic and equinox of date."""
    @staticmethod
    def rotation_at(t):
        return build_ecliptic_matrix(t)

ecliptic_frame = ecliptic_frame()

class InertialFrame(object):
    def __init__(self, doc, matrix):
        self.__doc__ = doc
        self._matrix = matrix

    def rotation_at(self, t):
        return self._matrix

equatorial_B1950_frame = InertialFrame(
    'Reference frame of the Earth’s mean equator and equinox at B1950.',
    _inertial_frames['B1950'],
)

ecliptic_J2000_frame = InertialFrame(
    'Reference frame of the true ecliptic and equinox at J2000.',
    _inertial_frames['ECLIPJ2000'],
)
galactic_frame = InertialFrame(
    'Galactic System II reference frame.',
    _inertial_frames['GALACTIC'],
)



"""Accuracy tests against data pulled from HORIZONS."""

import os
from numpy import max
from skyfield import api, framelib
from skyfield.api import Topos
from skyfield.constants import AU_M
from skyfield.trigonometry import position_angle_of

def _data_path(filename):
    return os.path.join(os.path.dirname(__file__), 'data', filename)

one_second = 1.0 / 24.0 / 60.0 / 60.0
arcsecond = 1.0 / 60.0 / 60.0
ra_arcsecond = 24.0 / 360.0 / 60.0 / 60.0
meter = 1.0 / AU_M

def ts():
    yield api.load.timescale()

def compare(value, expected_value, epsilon):
    if hasattr(value, 'shape') or hasattr(expected_value, 'shape'):
        assert max(abs(value - expected_value)) <= epsilon
    else:
        assert abs(value - expected_value) <= epsilon

def test_ecliptic_frame(ts):
    e = api.load('de421.bsp')
    jup = e['jupiter barycenter']
    astrometric = e['sun'].at(ts.utc(1980, 1, 1, 0, 0)).observe(jup)
    hlat, hlon, d = astrometric.ecliptic_latlon()
    compare(hlat.degrees, 1.013, 0.001)
    compare(hlon.degrees, 151.3229, 0.001)

def test_ecliptic_for_epoch_of_date(ts):
    e = api.load('de421.bsp')
    mars = e['mars barycenter']
    astrometric = e['earth'].at(ts.utc(1956, 1, 14, 6, 0, 0)).observe(mars)
    apparent = astrometric.apparent()
    hlat, hlon, d = apparent.ecliptic_latlon(epoch='date')
    compare(hlat.degrees, 0.4753402, 0.00001)
    compare(hlon.degrees, 240.0965633, 0.0002)

def test_ecliptic_for_epoch_of_date_array(ts):
    e = api.load('de421.bsp')
    sun = e['sun']
    astrometric = e['earth'].at(ts.utc(2005, 10, 1, [7, 8], 0, 0)).observe(sun)
    apparent = astrometric.apparent()
    hlat, hlon, d = apparent.ecliptic_latlon(epoch='date')
    compare(hlat.degrees[0], 0.0000488, 0.00001)
    compare(hlat.degrees[1], 0.0000474, 0.00001)
    compare(hlon.degrees[0], 188.2011122, 0.0002)
    compare(hlon.degrees[1], 188.2420983, 0.0002)

def test_fk4_frame(ts):
    e = api.load('de421.bsp')
    astrometric = e['earth'].at(ts.utc(1980, 1, 1, 0, 0)).observe(e['moon'])
    dec, ra, d = astrometric.frame_latlon(framelib.equatorial_B1950_frame)
    print(ra.degrees, dec.degrees)
    compare(ra.degrees, 82.36186, 0.00006) # TODO: why is this not 0.00001?
    compare(dec.degrees, 18.53432, 0.00006)

def test_galactic_frame(ts):
    e = api.load('de421.bsp')
    astrometric = e['earth'].at(ts.utc(1980, 1, 1, 0, 0)).observe(e['moon'])
    glat, glon, d = astrometric.frame_latlon(framelib.galactic_frame)
    print(glat, glat.degrees, glon, glon.degrees)
    compare(glat.degrees, -8.047315, 0.005)  # TODO: awful! Track this down.
    compare(glon.degrees, 187.221794, 0.005)

def test_callisto_geometry(ts):
    e = api.load_file(_data_path('jup310-2053-10-08.bsp'))
    a = (e['callisto'] - e['earth']).at(ts.tdb(jd=2471184.5))
    print(a)
    compare(a.xyz.au,
      [-4.884815926454119E+00, -3.705745549073268E+00, -1.493487818022234E+00],
      0.001 * meter)
    compare(a.velocity.au_per_d,
      [9.604665478763035E-03, -1.552997751083403E-02, -6.678445860769302E-03],
      0.00001 * meter)

def test_callisto_astrometric(ts):
    e = api.load_file(_data_path('jup310-2053-10-08.bsp'))
    # This date was utc(2053, 10, 9), but new leap seconds keep breaking
    # the test, so:
    a = e['earth'].at(ts.tt(jd=2471184.5007775929)).observe(e['callisto'])
    ra, dec, distance = a.radec()
    compare(ra.degrees, 217.1839292, 0.001 * arcsecond)
    compare(dec.degrees, -13.6892791, 0.001 * arcsecond)
    compare(distance.au, 6.31079291776184, 0.2 * meter)

def test_boston_geometry():
    e = api.load_file(_data_path('jup310-2015-03-02.bsp'))
    ts = api.load.timescale(delta_t=67.185390 + 0.5285957)
    ts.polar_motion_table = [0.0], [0.003483], [0.358609]
    t = ts.tdb(2015, 3, 2)
    boston = e['earth'] + Topos((42, 21, 24.1), (-71, 3, 24.8))
    a = (e['earth'] - boston).at(t)
    meter = 1e-3
    compare(a.xyz.km,
      [-1.764697476371664E+02, -4.717131288041386E+03, -4.274926422016179E+03],
      0.7 * meter)

def test_moon_from_boston_geometry():
    e = api.load_file(_data_path('de430-2015-03-02.bsp'))
    ts = api.load.timescale(delta_t=67.185390 + 0.5285957)
    ts.polar_motion_table = [0.0], [0.003483], [0.358609]
    t = ts.tdb(2015, 3, 2)
    boston = e['earth'] + Topos((42, 21, 24.1), (-71, 3, 24.8))
    a = (e['moon'] - boston).at(t)
    compare(a.xyz.au,
      [-1.341501206552443E-03, 2.190483327459023E-03, 6.839177007993498E-04],
      1.1 * meter)

def test_moon_from_boston_astrometric():
    e = api.load_file(_data_path('de430-2015-03-02.bsp'))
    ts = api.load.timescale(delta_t=67.185390 + 0.5285957)
    ts.polar_motion_table = [0.0], [0.003483], [0.358609]
    t = ts.tdb(2015, 3, 2)
    boston = e['earth'] + Topos((42, 21, 24.1), (-71, 3, 24.8))
    a = boston.at(t).observe(e['moon'])
    ra, dec, distance = a.radec()
    compare(ra.degrees, 121.4796470, 0.001 * arcsecond)
    compare(dec.degrees, 14.9108450, 0.001 * arcsecond)
    compare(distance.au, 0.00265828588792, 0.05 * meter)

def test_position_angle_from_boston(ts):
    t = ts.utc(2053, 10, 9)
    eph = api.load_file(_data_path('jup310-2053-10-08.bsp'))
    boston = eph['earth'] + Topos(longitude_degrees=(-71, 3, 24.8),
                                  latitude_degrees=(42, 21, 24.1))

    b = boston.at(t)
    j = b.observe(eph['jupiter'])#.apparent()
    i = b.observe(eph['io'])#.apparent()

    a = position_angle_of(j.radec(epoch='date'), i.radec(epoch='date'))

    assert abs(a.degrees - 293.671) < 0.002

    print(a)
    a = position_angle_of(j.ecliptic_latlon(epoch='date'),
                          i.ecliptic_latlon(epoch='date'))
    print(a)
    #TODO: fix misbehavior for ecliptic coordinates that is illustrated above



"""Test whether Skyfield handles file download and file age correctly."""

import os
import shutil
import tempfile
from contextlib import contextmanager
from threading import Thread
try:
    from Queue import Queue
except ImportError:
    from queue import Queue
try:
    from unittest.mock import patch
except ImportError:
    from mock import patch

from assay import assert_raises
from skyfield import api

leap_second_text = (
    b'#  File expires on 28 June 2021\n'
    b'    41317.0    1  1 1972       10\n'
    b'    41499.0    1  7 1972       11\n'
)

def load():
    path = tempfile.mkdtemp()
    try:
        yield api.Loader(path)
    finally:
        shutil.rmtree(path)

def save_file(load, path, content):
    with open(load.path_to(path), 'wb') as f:
        f.write(content)

def file_contents(load, path):
    with open(load.path_to(path), 'rb') as f:
        return f.read()

@contextmanager
def fake_download(load, filename, content):
    download = lambda *args, **kw: save_file(load, filename, content)
    with patch('skyfield.iokit.download', download):
        yield

# Simple tests.

def test_build_url(load):
    url = 'https://ssd.jpl.nasa.gov/ftp/eph/planets/bsp/de421.bsp'
    assert load.build_url('de421.bsp') == url
    with assert_raises(ValueError, 'know the URL'):
        load.build_url('unknown.kind.of.file')

def test_open_in_main_directory(load):
    with open(os.path.join(load.directory, 'file.tle'), 'wb') as f:
        f.write(b'example text\n')
    data = load.open('file.tle').read()
    assert data == b'example text\n'

def test_open_in_subdirectory(load):
    os.mkdir(os.path.join(load.directory, 'folder'))
    with open(os.path.join(load.directory, 'folder', 'file.tle'), 'wb') as f:
        f.write(b'example text\n')
    data = load.open('folder/file.tle').read()
    assert data == b'example text\n'

def test_missing_file_gets_downloaded(load):
    with fake_download(load, 'Leap_Second.dat', leap_second_text):
        data = load('Leap_Second.dat')
    assert file_contents(load, 'Leap_Second.dat') == leap_second_text
    assert list(data[1][1]) == [10, 10, 10, 11]

def test_builtin_timescale_uses_recent_IERS_data(load):
    ts = load.timescale()
    # DUT1 cut and pasted from "20 1 1" line of "finals2000A.all":
    assert abs(ts.utc(2020, 1, 1).dut1 - (-0.1771554)) < 1e-8

def test_non_builtin_timescale_prefers_USNO_files(load):
    save_file(load, 'deltat.data',
              b' 1973  2  1  111.1\n'
              b' 1973  3  1  222.2\n'
              b' 1973  4  1  333.3\n')
    save_file(load, 'deltat.preds',
              b'   MJD        YEAR    TT-UT Pred  UT1-UTC Pred  ERROR\n'
              b'   58484.000  2019.00   69.34      -0.152       0.117\n'
              b'   58575.000  2019.25   69.48      -0.295       0.162\n')
    save_file(load, 'Leap_Second.dat', leap_second_text)
    save_file(load, 'finals2000A.all', b'invalid data')

    ts = load.timescale(builtin=False)
    assert abs(ts.utc(1973, 3, 1).delta_t - 222.2) < 1e-2

    # Did we correctly convert the old leap second table to the new
    # format? The offset prior to the first leap second should be +10.
    t = ts.tai(1970, 1, 1)
    assert t.utc == (1969, 12, 31, 23, 59, 50)

def test_non_builtin_timescale_tries_to_load_finals2000A_all(load):
    save_file(load, 'finals2000A.all', b'invalid data')
    with assert_raises(IndexError):
        load.timescale(builtin=False)

def test_close_of_ephemeris(load):
    eph = api.load('de421.bsp')
    assert eph.segments
    assert eph.codes
    eph.close()
    assert eph.segments == []
    assert eph.codes == set()

# Impressive tests: synchronize threads to reproduce concurrency bugs.

class FakeConnection():
    def __init__(self):
        self.control = Queue()
        self.blocks = Queue()
        self.headers = {}

    def __call__(self, *args, **kw):  # Intercept urlopen()
        return self

    def read(self, blocksize):  # Intercept connection.read()
        self.control.put('Read requested')
        return self.blocks.get()

def test_concurrent_downloads(load):
    # While downloads in a single process are protecte

    eof = b''
    data = b' 1973  2  1  43.4724\n'

    c1 = FakeConnection()
    c2 = FakeConnection()
    t1 = Thread(target=load, args=('Leap_Second.dat',))
    t2 = Thread(target=load, args=('Leap_Second.dat',))
    t1.daemon = True
    t2.daemon = True

    with patch('skyfield.iokit.urlopen', c1):
        t1.start()
        c1.control.get()  # has reached its first read()
    with patch('skyfield.iokit.urlopen', c2):
        t2.start()
        c2.control.get()  # has reached its first read()

    assert sorted(os.listdir(load.directory)) == [
        'Leap_Second.dat.download', 'Leap_Second.dat.download2',
    ]
    c1.blocks.put(data)
    c1.blocks.put(eof)
    t1.join()
    assert sorted(os.listdir(load.directory)) == [
        'Leap_Second.dat', 'Leap_Second.dat.download2',
    ]
    c2.blocks.put(data)
    c2.blocks.put(eof)
    t2.join()
    assert sorted(os.listdir(load.directory)) == ['Leap_Second.dat']



"""Tests of the Skyfield `api` module and user-facing exceptions."""

import numpy as np
from assay import assert_raises
from skyfield import api, positionlib
from skyfield.api import Topos
from skyfield.errors import EphemerisRangeError

def ts():
    yield api.load.timescale()

def test_sending_jd_that_is_not_a_julian_date():
    earth = api.load('de421.bsp')['earth']
    with assert_raises(ValueError, r"please provide the at\(\) method"
                       " with a Time instance as its argument,"
                       " instead of the value 'blah'"):
        earth.at('blah')

def test_observe_and_apparent_survive_an_empty_time_array(ts):
    e = api.load('de421.bsp')
    e['earth'].at(ts.utc(2024, 2, 4, 7, [])).observe(e['mars']).apparent()

def test_apparent_position_class(ts):
    e = api.load('de421.bsp')
    p = e['earth'].at(ts.utc(2014, 2, 9, 14, 50)).observe(e['mars']).apparent()
    assert isinstance(p, positionlib.Apparent)

def test_astrometric_position_class(ts):
    e = api.load('de421.bsp')
    p = e['earth'].at(ts.utc(2014, 2, 9, 14, 50)).observe(e['mars'])
    assert isinstance(p, positionlib.Astrometric)

def test_astrometric_position_does_not_allow_altaz(ts):
    e = api.load('de421.bsp')
    o = e['earth'] + api.wgs84.latlon(36.7138, -112.2169)
    a = o.at(ts.utc(2014, 2, 9, 14, 50)).observe(e['mars'])
    with assert_raises(ValueError, 'it is not useful to call .altaz'):
        a.altaz()

def test_ephemeris_in_operator_accepts_strings(ts):
    e = api.load('de421.bsp')
    assert ('earth' in e) is True
    assert ('Earth' in e) is True
    assert ('EARTH' in e) is True
    assert ('ceres' in e) is False

def test_exception_raised_for_dates_outside_ephemeris(ts):
    eph = api.load('de421.bsp')
    message = (
        'ephemeris segment only covers dates 1899-07-29 through 2053-10-09'
    )
    with assert_raises(EphemerisRangeError, message) as a:
        eph['earth'].at(ts.tt(4096))

    e = a.exception
    assert e.args == (message,)
    assert e.start_time.tdb == 2414864.5
    assert e.end_time.tdb == 2471184.5
    assert e.time_mask == [True]
    assert e.segment is eph['earth'].vector_functions[0].spk_segment

def test_planet_position_class(ts):
    e = api.load('de421.bsp')
    p = e['mars'].at(ts.utc(2014, 2, 9, 14, 50))
    assert isinstance(p, positionlib.Barycentric)

def test_star_position_class(ts):
    e = api.load('de421.bsp')
    star = api.Star(ra_hours=0, dec_degrees=0)
    p = e['earth'].at(ts.utc(2014, 2, 9, 15, 1)).observe(star)
    assert isinstance(p, positionlib.Astrometric)

def test_star_vector_from_earth(ts):
    t = ts.tt_jd(api.T0)
    eph = api.load('de421.bsp')
    e = eph['earth'].at(t)

    star = api.Star(ra_hours=[1.0, 2.0], dec_degrees=[+3.0, +4.0])
    p = e.observe(star)
    assert p.xyz.au.shape == (3, 2)
    assert p.velocity.au_per_d.shape == (3, 2)
    assert p.t.shape == (2,)
    assert (p.t.tt == api.T0).all()
    a = p.apparent()

    a1 = e.observe(api.Star(ra_hours=1.0, dec_degrees=+3.0)).apparent()
    a2 = e.observe(api.Star(ra_hours=2.0, dec_degrees=+4.0)).apparent()
    assert (a1.xyz.au == a.xyz.au[:,0]).all()
    assert (a2.xyz.au == a.xyz.au[:,1]).all()

def test_star_vector_from_topos(ts):
    t = ts.tt_jd(api.T0)
    eph = api.load('de421.bsp')
    boston = eph['earth'] + Topos('42.3583 N', '71.0636 W')
    b = boston.at(t)

    star = api.Star(ra_hours=[1.0, 2.0], dec_degrees=[+3.0, +4.0])
    p = b.observe(star)
    assert p.xyz.au.shape == (3, 2)
    assert p.velocity.au_per_d.shape == (3, 2)
    assert p.t.shape == (2,)
    assert (p.t.tt == api.T0).all()
    a = p.apparent()

    a1 = b.observe(api.Star(ra_hours=1.0, dec_degrees=+3.0)).apparent()
    a2 = b.observe(api.Star(ra_hours=2.0, dec_degrees=+4.0)).apparent()
    assert (a1.xyz.au == a.xyz.au[:,0]).all()
    assert (a2.xyz.au == a.xyz.au[:,1]).all()

def test_hadec_needs_a_longitude(ts):
    e = api.load('de421.bsp')
    earth = e['earth']
    moon = e['moon']
    apparent = earth.at(ts.utc(2016)).observe(moon)
    with assert_raises(ValueError, 'from a specific latitude and longitude'):
        apparent.hadec()

def test_altaz_needs_topos(ts):
    e = api.load('de421.bsp')
    earth = e['earth']
    moon = e['moon']
    apparent = earth.at(ts.utc(2016)).observe(moon).apparent()
    with assert_raises(ValueError, 'from a specific latitude and longitude'):
        apparent.altaz()

def test_from_altaz_needs_topos():
    p = positionlib.ICRF([0.0, 0.0, 0.0])
    with assert_raises(ValueError, 'to compute altitude and azimuth'):
        p.from_altaz(alt_degrees=0, az_degrees=0)

def test_from_altaz_parameters(ts):
    usno = Topos('38.9215 N', '77.0669 W', elevation_m=92.0)
    t = ts.tt(jd=api.T0)
    p = usno.at(t)
    a = api.Angle(degrees=10.0)
    d = api.Distance(au=0.234)
    with assert_raises(ValueError, 'the alt= parameter with an Angle'):
        p.from_altaz(alt='Bad value', alt_degrees=0, az_degrees=0)
    with assert_raises(ValueError, 'the az= parameter with an Angle'):
        p.from_altaz(az='Bad value', alt_degrees=0, az_degrees=0)
    p.from_altaz(alt=a, alt_degrees='bad', az_degrees=0)
    p.from_altaz(az=a, alt_degrees=0, az_degrees='bad')
    assert str(p.from_altaz(alt=a, az=a).distance()) == '0.1 au'
    assert str(p.from_altaz(alt=a, az=a, distance=d).distance()) == '0.234 au'

def test_github_81(ts):
    t = ts.utc(1980, 1, 1)
    assert t == t
    assert (t == 61) is False  # used to die with AttributeError for "tt"

def test_github_500_does_zero_position_trigger_numpy_warnings(ts):
    zero_vector = np.array([0.0, 0.0, 0.0])
    t = ts.utc(2020, 12, 14)
    p = positionlib.Astrometric(zero_vector, zero_vector, t=t, center=0)
    p._ephemeris = api.load('de421.bsp')
    with np.errstate(all='raise'):
        p.apparent().radec()



# -*- coding: utf-8 -*-
"""Simple distance, velocity, and angle support for Skyfield.

"""
import numpy as np
from numpy import abs, copysign, isnan
from types import MethodType
from .constants import AU_KM, AU_M, C, DAY_S, tau
from .descriptorlib import reify
from .functions import _to_array, length_of

_dfmt = '{0}{1:02}deg {2:02}\' {3:02}.{4:0{5}}"'
_dsgn = '{0:+>1}{1:02}deg {2:02}\' {3:02}.{4:0{5}}"'
_hfmt = '{0}{1:02}h {2:02}m {3:02}.{4:0{5}}s'

class UnpackingError(Exception):
    """You cannot iterate directly over a Skyfield measurement object."""

class Unit(object):
    """A measurement that can be expressed in several choices of unit."""

    def __getitem__(self, *args):
        """Tell users to ask for a specific unit before indexing or slicing."""
        cls = self.__class__
        name = cls.__name__
        s = 'to use this {0}, ask for its value in a particular unit:\n\n{1}'
        attrs = sorted(k for k, v in cls.__dict__.items()
                       if k[0].islower() and isinstance(v, (getset, reify)))
        examples = ['    {0}.{1}'.format(name.lower(), attr) for attr in attrs]
        raise UnpackingError(s.format(name, '\n'.join(examples)))

    __iter__ = __getitem__   # give advice about both foo[i] and "x,y,z = foo"

class getset(object):
    """Unit name that serves as both a class constructor and instance attribute.

    This supports two use cases:

    * When called as a class method like ``Distance.km(5.0)``, we build
      and return an instance of ``Distance`` whose ``km`` has been set
      to 5.0 and whose base unit ``m``, using the appropriate conversion
      factor, has been set to 5000.0.

    * When invoked like ``d.km`` on a particular ``Distance`` that
      doesn't yet have a ``km`` attribute (which otherwise Python itself
      would have returned), we apply the conversion factor to ``d.m``
      and return the result.

    """
    def __init__(self, name, docstring, conversion_factor=None, core_unit=None):
        self.name = name
        self.__doc__ = docstring
        self.conversion_factor = conversion_factor
        self.core_unit = core_unit

        if conversion_factor is None:
            def _constructor(cls, value):
                value = _to_array(value)
                obj = cls.__new__(cls)
                setattr(obj, name, value)
                return obj

        else:
            def _constructor(cls, value):
                value = _to_array(value)
                obj = cls.__new__(cls)
                setattr(obj, name, value)
                setattr(obj, core_unit, value / conversion_factor)
                return obj

        _constructor.__doc__ = self.__doc__
        self._constructor = _constructor

    def __get__(self, instance, objtype=None):
        if instance is None:  # the class itself has been asked for this name
            return MethodType(self._constructor, objtype)
        value = getattr(instance, self.core_unit) * self.conversion_factor
        instance.__dict__[self.name] = value
        return value

class Distance(Unit):
    """A distance, stored internally as au and available in other units.

    You can initialize a ``Distance`` by providing a single float or a
    float array as either an ``au=``, ``km=``, or ``m=`` parameter.

    You can access the magnitude of the distance with its three
    attributes ``.au``, ``.km``, and ``.m``.  By default a distance
    prints itself in astronomical units (au), but you can take control
    of the formatting and choice of units yourself using standard Python
    numeric formatting:

    >>> d = Distance(au=1)
    >>> print(d)
    1.0 au
    >>> print('{:.2f} km'.format(d.km))
    149597870.70 km

    """
    def __init__(self, au=None, km=None, m=None):
        if au is not None:
            self.au = _to_array(au)
        elif km is not None:
            self.km = km = _to_array(km)
            self.au = km / AU_KM
        elif m is not None:
            self.m = m = _to_array(m)
            self.au = m / AU_M
        else:
            raise ValueError('to construct a Distance provide au, km, or m')

    @classmethod
    def from_au(cls, au):  # deprecated and no longer used internally
        return cls.au(au)

    au = getset('au', 'Astronomical units'
                ' (the Earth-Sun distance of 149,597,870,700 m).')
    km = getset('km', 'Kilometers (1,000 meters).', AU_KM, 'au')
    m = getset('m', 'Meters.', AU_M, 'au')

    def __str__(self):
        n = self.au
        return ('{0} au' if getattr(n, 'shape', 0) else '{0:.6} au').format(n)

    def __repr__(self):
        return '<{0} {1}>'.format(type(self).__name__, self)

    def length(self):
        """Compute the length when this is an |xyz| vector.

        The Euclidean vector length of this vector is returned as a new
        :class:`~skyfield.units.Distance` object.

        >>> from skyfield.api import Distance
        >>> d = Distance(au=[1, 1, 0])
        >>> d.length()
        <Distance 1.41421 au>

        """
        return Distance(au=length_of(self.au))

    def light_seconds(self):
        """Return the length of this vector in light seconds."""
        return self.m / C

    def to(self, unit):
        """Convert this distance to the given AstroPy unit."""
        from astropy.units import au
        return (self.au * au).to(unit)

class Velocity(Unit):
    """A velocity, stored internally as au/day and available in other units.

    You can initialize a ``Velocity`` by providing a float or float
    array to its ``au_per_d=`` parameter.

    """
    # TODO: consider reworking this class to return a Rate object.

    def __init__(self, au_per_d=None, km_per_s=None):
        if km_per_s is not None:
            self.km_per_s = km_per_s = _to_array(km_per_s)
            self.au_per_d = km_per_s * DAY_S / AU_KM
        elif au_per_d is not None:
            self.au_per_d = _to_array(au_per_d)
        else:
            raise ValueError('to construct a Velocity provide'
                             ' au_per_d or km_per_s')

    au_per_d = getset('au_per_d', 'Astronomical units per day.')
    km_per_s = getset('km_per_s', 'Kilometers per second.',
                      AU_KM / DAY_S, 'au_per_d')
    m_per_s = getset('m_per_s', 'Meters per second.',
                     AU_M / DAY_S, 'au_per_d')

    def __str__(self):
        n = self.au_per_d
        fmt = '{0} au/day' if getattr(n, 'shape', 0) else '{0:.6} au/day'
        return fmt.format(n)

    def __repr__(self):
        return '<{0} {1}>'.format(type(self).__name__, self)

    def to(self, unit):
        """Convert this velocity to the given AstroPy unit."""
        from astropy.units import au, d
        return (self.au_per_d * au / d).to(unit)

class AngleRate(object):
    """The rate at which an angle is changing."""

    # TODO: design and implement public constructor.

    @classmethod
    def _from_radians_per_day(cls, radians_per_day):
        ar = cls()
        ar._radians_per_day = radians_per_day
        return ar

    @reify
    def radians(self):
        """:class:`Rate` of change in radians."""
        return Rate._from_per_day(self._radians_per_day)

    @reify
    def degrees(self):
        """:class:`Rate` of change in degrees."""
        return Rate._from_per_day(self._radians_per_day / tau * 360.0)

    @reify
    def arcminutes(self):
        """:class:`Rate` of change in arcminutes."""
        return Rate._from_per_day(self._radians_per_day / tau * 21600.0)

    @reify
    def arcseconds(self):
        """:class:`Rate` of change in arcseconds."""
        return Rate._from_per_day(self._radians_per_day / tau * 1296000.0)

    @reify
    def mas(self):
        """:class:`Rate` of change in milliarcseconds."""
        return Rate._from_per_day(self._radians_per_day / tau * 1.296e9)

    # TODO: str; repr; conversion to AstroPy units

class Rate(object):
    """Measurement whose denominator is time."""

    # TODO: design and implement public constructor.

    @classmethod
    def _from_per_day(cls, per_day):
        r = cls()
        r._per_day = per_day
        return r

    @reify
    def per_day(self):
        """Units per day of Terrestrial Time."""
        return self._per_day

    @reify
    def per_hour(self):
        """Units per hour of Terrestrial Time."""
        return self._per_day / 24.0

    @reify
    def per_minute(self):
        """Units per minute of Terrestrial Time."""
        return self._per_day / 1440.0

    @reify
    def per_second(self):
        """Units per second of Terrestrial Time."""
        return self._per_day / 86400.0

# Angle units.

_instantiation_instructions = """to instantiate an Angle, try one of:

Angle(angle=another_angle)
Angle(radians=value)
Angle(degrees=value)
Angle(hours=value)

where `value` can be either a Python float, a list of Python floats,
or a NumPy array of floats"""

class Angle(Unit):

    def __init__(self, angle=None, radians=None, degrees=None, hours=None,
                 preference=None, signed=False):

        if angle is not None:
            if not isinstance(angle, Angle):
                raise ValueError(_instantiation_instructions)
            self.radians = angle.radians
        elif radians is not None:
            self.radians = _to_array(radians)
        elif degrees is not None:
            self.degrees = degrees = _to_array(_unsexagesimalize(degrees))
            self.radians = degrees / 360.0 * tau
        elif hours is not None:
            self.hours = hours = _to_array(_unsexagesimalize(hours))
            self.radians = hours / 24.0 * tau

        self.preference = (preference if preference is not None
                           else 'hours' if hours is not None
                           else 'degrees')
        self.signed = signed

    @classmethod
    def from_degrees(cls, degrees, signed=False):
        degrees = _to_array(_unsexagesimalize(degrees))
        self = cls.__new__(cls)
        self.degrees = degrees
        self.radians = degrees / 360.0 * tau
        self.preference = 'degrees'
        self.signed = signed
        return self

    radians = getset('radians', 'Radians (𝜏 = 2𝜋 in a circle).')

    # Deprecated names, to support legacy code.
    @reify
    def _hours(self): return self.hours
    @reify
    def _degrees(self): return self.degrees

    @reify
    def hours(self):
        r"""Hours (24\ |h| in a circle)."""
        return self.radians * 24.0 / tau

    @reify
    def degrees(self):
        """Degrees (360° in a circle)."""
        return self.radians * 360.0 / tau

    def arcminutes(self):
        """Return the angle in arcminutes."""
        return self.degrees * 60.0

    def arcseconds(self):
        """Return the angle in arcseconds."""
        return self.degrees * 3600.0

    def mas(self):
        """Return the angle in milliarcseconds."""
        return self.degrees * 3600000.0

    def __str__(self):
        size = self.radians.size
        if size == 0:
            return 'Angle []'
        if self.preference == 'degrees':
            v = self.degrees
            fmt = _dsgn.format if self.signed else _dfmt.format
            places = 1
        else:
            v = self.hours
            fmt = _hfmt.format
            places = 2
        if size >= 2:
            return '{0} values from {1} to {2}'.format(
                len(v), _sfmt(fmt, v[0], places), _sfmt(fmt, v[-1], places))
        return _sfmt(fmt, v, places)

    def __repr__(self):
        if self.radians.size == 0:
            return '<{0} []>'.format(type(self).__name__)
        else:
            return '<{0} {1}>'.format(type(self).__name__, self)

    def hms(self, warn=True):
        """Convert to a tuple (hours, minutes, seconds).

        All three quantities will have the same sign as the angle itself.

        """
        if warn and self.preference != 'hours':
            raise WrongUnitError('hms')
        sign, units, minutes, seconds = _sexagesimalize_to_float(self.hours)
        return sign * units, sign * minutes, sign * seconds

    def signed_hms(self, warn=True):
        """Convert to a tuple (sign, hours, minutes, seconds).

        The ``sign`` will be either +1 or -1, and the other quantities
        will all be positive.

        """
        if warn and self.preference != 'hours':
            raise WrongUnitError('signed_hms')
        return _sexagesimalize_to_float(self.hours)

    def hstr(self, places=2, warn=True, format=_hfmt):
        """Return a string like ``12h 07m 30.00s``; see `Formatting angles`.

        .. versionadded:: 1.39

           Added the ``format=`` parameter.

        """
        if warn and self.preference != 'hours':
            raise WrongUnitError('hstr')
        hours = self.hours
        shape = getattr(hours, 'shape', ())
        fmt = format.format  # `format()` method of `format` string
        if shape:
            return [_sfmt(fmt, h, places) for h in hours]
        return _sfmt(fmt, hours, places)

    def dms(self, warn=True):
        """Convert to a tuple (degrees, minutes, seconds).

        All three quantities will have the same sign as the angle itself.

        """
        if warn and self.preference != 'degrees':
            raise WrongUnitError('dms')
        sign, units, minutes, seconds = _sexagesimalize_to_float(self.degrees)
        return sign * units, sign * minutes, sign * seconds

    def signed_dms(self, warn=True):
        """Convert to a tuple (sign, degrees, minutes, seconds).

        The ``sign`` will be either +1 or -1, and the other quantities
        will all be positive.

        """
        if warn and self.preference != 'degrees':
            raise WrongUnitError('signed_dms')
        return _sexagesimalize_to_float(self.degrees)

    def dstr(self, places=1, warn=True, format=None):
        """Return a string like ``181deg 52' 30.0"``; see `Formatting angles`.

        .. versionadded:: 1.39

           Added the ``format=`` parameter.

        """
        if warn and self.preference != 'degrees':
            raise WrongUnitError('dstr')
        degrees = self.degrees
        signed = self.signed
        if format is None:
            format = _dsgn if signed else _dfmt
        fmt = format.format  # `format()` method of `format` string
        shape = getattr(degrees, 'shape', ())
        if shape:
            return [_sfmt(fmt, d, places) for d in degrees]
        return _sfmt(fmt, degrees, places)

    def to(self, unit):
        """Convert this angle to the given AstroPy unit."""
        from astropy.units import rad
        return (self.radians * rad).to(unit)

        # Or should this do:
        from astropy.coordinates import Angle
        from astropy.units import rad
        return Angle(self.radians, rad).to(unit)

class WrongUnitError(ValueError):

    def __init__(self, name):
        unit = 'hours' if (name.startswith('h') or '_h' in name) else 'degrees'
        usual = 'hours' if (unit == 'degrees') else 'degrees'
        message = ('this angle is usually expressed in {0}, not {1};'
                   ' if you want to use {1} anyway,'.format(usual, unit))
        if name == unit:
            message += ' then please use the attribute _{0}'.format(unit)
        else:
            message += ' then call {0}() with warn=False'.format(name)
        self.args = (message,)

def _sexagesimalize_to_float(value):
    """Decompose `value` into units, minutes, and seconds.

    Note that this routine is not appropriate for displaying a value,
    because rounding to the smallest digit of display is necessary
    before showing a value to the user.  Use `_sexagesimalize_to_int()`
    for data being displayed to the user.

    This routine simply decomposes the floating point `value` into a
    sign (+1.0 or -1.0), units, minutes, and seconds, returning the
    result in a four-element tuple.

    >>> _sexagesimalize_to_float(12.05125)
    (1.0, 12.0, 3.0, 4.5)
    >>> _sexagesimalize_to_float(-12.05125)
    (-1.0, 12.0, 3.0, 4.5)

    """
    sign = np.sign(value)
    n = abs(value)
    minutes, seconds = divmod(n * 3600.0, 60.0)
    units, minutes = divmod(minutes, 60.0)
    return sign, units, minutes, seconds

def _sexagesimalize_to_int(value, places=0):
    """Decompose `value` into units, minutes, seconds, and second fractions.

    This routine prepares a value for sexagesimal display, with its
    seconds fraction expressed as an integer with `places` digits.  The
    result is a tuple of five integers:

    ``(sign [either +1 or -1], units, minutes, seconds, second_fractions)``

    The integers are properly rounded per astronomical convention so
    that, for example, given ``places=3`` the result tuple ``(1, 11, 22,
    33, 444)`` means that the input was closer to 11u 22' 33.444" than
    to either 33.443" or 33.445" in its value.

    """
    power = 10 ** places
    n = int((power * 3600 * value + 0.5) // 1.0)
    sign = np.sign(n)
    n, fraction = divmod(abs(n), power)
    n, seconds = divmod(n, 60)
    n, minutes = divmod(n, 60)
    return sign, n, minutes, seconds, fraction

def _sfmt(fmt, value, places):
    """Decompose floating point `value` into sexagesimal, and format."""
    if isnan(value):
        return 'nan'
    sgn, h, m, s, fraction = _sexagesimalize_to_int(value, places)
    sign = '-' if sgn < 0.0 else ''
    return fmt(sign, h, m, s, fraction, places)

def wms(whole, minutes=0.0, seconds=0.0):
    """Return a quantity expressed with 1/60 minutes and 1/3600 seconds."""
    return (whole
            + copysign(minutes, whole) / 60.0
            + copysign(seconds, whole) / 3600.0)

def _unsexagesimalize(value):
    """Return `value` after interpreting a (units, minutes, seconds) tuple.

    When `value` is not a tuple, it is simply returned.

    >>> _unsexagesimalize(3.25)
    3.25

    An input tuple is interpreted as units, minutes, and seconds.  Note
    that only the sign of `units` is significant!  So all of the
    following tuples convert into exactly the same value:

    >>> '%f' % _unsexagesimalize((-1, 2, 3))
    '-1.034167'
    >>> '%f' % _unsexagesimalize((-1, -2, 3))
    '-1.034167'
    >>> '%f' % _unsexagesimalize((-1, -2, -3))
    '-1.034167'

    """
    if isinstance(value, tuple):
        components = iter(value)
        value = next(components)
        factor = 1.0
        for component in components:
            factor *= 60.0
            value += copysign(component, value) / factor
    return value

def _interpret_angle(name, angle_object, angle_float, unit='degrees'):
    """Return an angle in radians from one of two arguments.

    It is common for Skyfield routines to accept both an argument like
    `alt` that takes an Angle object as well as an `alt_degrees` that
    can be given a bare float or a sexagesimal tuple.  A pair of such
    arguments can be passed to this routine for interpretation.

    """
    if angle_object is not None:
        if isinstance(angle_object, Angle):
            return angle_object.radians
    elif angle_float is not None:
        return _unsexagesimalize(angle_float) / 360.0 * tau
    raise ValueError('you must either provide the {0}= parameter with'
                     ' an Angle argument or supply the {0}_{1}= parameter'
                     ' with a numeric argument'.format(name, unit))

def _ltude(value, name, psuffix, nsuffix):
    # Support for old deprecated Topos argument interpretation.
    if not isinstance(value, str):
        return _unsexagesimalize(value)
    value = value.strip().upper()
    if value.endswith(psuffix):
        sign = +1.0
    elif value.endswith(nsuffix):
        sign = -1.0
    else:
        raise ValueError('your {0} string {1!r} does not end with either {2!r}'
                         ' or {3!r}'.format(name, value, psuffix, nsuffix))
    try:
        value = float(value[:-1])
    except ValueError:
        raise ValueError('your {0} string {1!r} cannot be parsed as a floating'
                         ' point number'.format(name, value))
    return sign * value



from numpy import array, concatenate as concat
from skyfield import api, almanac

# Compare with USNO:
# http://aa.usno.navy.mil/cgi-bin/aa_moonill2.pl?form=1&year=2018&task=00&tz=-05

def test_fraction_illuminated():
    ts = api.load.timescale()
    t0 = ts.utc(2018, 9, range(9, 19), 5)
    e = api.load('de421.bsp')
    i = almanac.fraction_illuminated(e, 'moon', t0[-1]).round(2)
    assert i == 0.62
    i = almanac.fraction_illuminated(e, 'moon', t0).round(2)
    assert list(i) == [0, 0, 0.03, 0.08, 0.15, 0.24, 0.33, 0.43, 0.52, 0.62]

# Compare with USNO:
# http://aa.usno.navy.mil/seasons?year=2018&tz=+0

def test_seasons():
    ts = api.load.timescale()
    t0 = ts.utc(2018, 9, 20)
    t1 = ts.utc(2018, 12, 31)
    e = api.load('de421.bsp')
    t, y = almanac.find_discrete(t0, t1, almanac.seasons(e))
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2018-09-23 01:54', '2018-12-21 22:23']
    assert (y == (2, 3)).all()

# Compare with USNO:
# http://aa.usno.navy.mil/cgi-bin/aa_phases.pl?year=2018&month=9&day=11&nump=50&format=p

def test_moon_phases():
    ts = api.load.timescale()
    t0 = ts.utc(2018, 9, 11)
    t1 = ts.utc(2018, 9, 30)
    e = api.load('de421.bsp')
    t, y = almanac.find_discrete(t0, t1, almanac.moon_phases(e))
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2018-09-16 23:15', '2018-09-25 02:52']
    assert (y == (1, 2)).all()

def test_moon_nodes():
    ts = api.load.timescale()
    e = api.load('de421.bsp')
    t0 = ts.utc(2022, 1, 13)
    t1 = ts.utc(2022, 1, 31)
    t, y = almanac.find_discrete(t0, t1, almanac.moon_nodes(e))
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2022-01-13 04:19', '2022-01-27 06:15']
    assert list(y) == [1, 0]

def test_oppositions_conjunctions():
    ts = api.load.timescale()
    t0 = ts.utc(2019, 1, 1)
    t1 = ts.utc(2021, 1, 1)
    e = api.load('de421.bsp')
    f = almanac.oppositions_conjunctions(e, e['mars'])
    t, y = almanac.find_discrete(t0, t1, f)
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2019-09-02 10:42', '2020-10-13 23:26']
    assert (y == (0, 1)).all()

def test_oppositions_conjunctions_of_moon():
    ts = api.load.timescale()
    t0 = ts.utc(2019, 1, 1)
    t1 = ts.utc(2019, 2, 1)
    e = api.load('de421.bsp')
    f = almanac.oppositions_conjunctions(e, e['moon'])
    t, y = almanac.find_discrete(t0, t1, f)
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2019-01-06 01:28', '2019-01-21 05:16']
    assert (y == (1, 0)).all()

# Compare with USNO:
# http://aa.usno.navy.mil/rstt/onedaytable?ID=AA&year=2018&month=9&day=12&state=OH&place=Bluffton

def _sunrise_sunset(f):
    ts = api.load.timescale()
    t0 = ts.utc(2018, 9, 12, 4)
    t1 = ts.utc(2018, 9, 13, 4)
    e = api.load('de421.bsp')
    bluffton = api.Topos('40.8939 N', '83.8917 W')
    t, y = f(t0, t1, e, bluffton)
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2018-09-12 11:13', '2018-09-12 23:50']
    assert (y == (1, 0)).all()

def _concat(t0, t1):
    return t0.ts.tt_jd(concat((t0.whole, t1.whole)),
                       concat((t0.tt_fraction, t1.tt_fraction)))

def test_sunrise_sunset_old():
    def f(t0, t1, e, topos):
        return almanac.find_discrete(t0, t1, almanac.sunrise_sunset(e, topos))
    _sunrise_sunset(f)

def test_sunrise_sunset_new():
    def f(t0, t1, e, topos):
        r, _ = almanac.find_risings(e['earth'] + topos, e['sun'], t0, t1)
        s, _ = almanac.find_settings(e['earth'] + topos, e['sun'], t0, t1)
        t = _concat(r, s)
        return t, array([1, 0])
    _sunrise_sunset(f)

def test_new_moonrise_at_70_degrees_north():
    ts = api.load.timescale()
    t0 = ts.utc(2023, 2, 19)
    t1 = ts.utc(2023, 2, 20)
    e = api.load('de421.bsp')
    seventy = e['earth'] + api.wgs84.latlon(70, 0)

    t, y = almanac.find_risings(seventy, e['moon'], t0, t1)
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2023-02-19 11:28']
    assert list(y) == [1]

    t, y = almanac.find_settings(seventy, e['moon'], t0, t1)
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2023-02-19 12:05']
    assert list(y) == [1]

def test_dark_twilight_day():
    ts = api.load.timescale()
    t0 = ts.utc(2019, 11, 8, 4)
    t1 = ts.utc(2019, 11, 9, 4)
    e = api.load('de421.bsp')
    defiance = api.Topos('41.281944 N', '84.362778 W')
    t, y = almanac.find_discrete(t0, t1, almanac.dark_twilight_day(e, defiance))
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == [
        '2019-11-08 10:42', '2019-11-08 11:15', '2019-11-08 11:48',
        '2019-11-08 12:17', '2019-11-08 22:25', '2019-11-08 22:54',
        '2019-11-08 23:27', '2019-11-08 23:59',
    ]
    assert (y == (1, 2, 3, 4, 3, 2, 1, 0)).all()

# Logic.

def test_close_start_and_end():
    ts = api.load.timescale()
    t0 = ts.utc(2018, 9, 23, 1)
    t1 = ts.utc(2018, 9, 23, 2)
    e = api.load('de421.bsp')
    t, y = almanac.find_discrete(t0, t1, almanac.seasons(e))
    strings = t.utc_strftime('%Y-%m-%d %H:%M')
    assert strings == ['2018-09-23 01:54']



import numpy as np

# Old versions of interp() sometimes return a plain Python float,
# instead of always returning at least a NumPy float64 object.

interp = np.interp

if isinstance(interp(1.0, [0.0, 2.0], [10.0, 20.0]), float):
    def interp(*args):
        return np.array(np.interp(*args))



from numpy import array
from skyfield.data import iers
from skyfield.timelib import Timescale

def test_build_timescale_arrays():
    mjd = array([42046.00, 42047.00, 42048.00, 42049.00])
    dut1 = array([-0.2942581, -0.2971424, 0.6999438, 0.6970539])
    daily_tt, daily_delta_t, leap_dates, leap_offsets = (
        iers.build_timescale_arrays(mjd, dut1))

    assert list(daily_tt) == [2442046.5005113888, 2442047.5005113888,
                              2442048.500522963, 2442049.500522963]
    assert list(daily_delta_t) == [44.4782581, 44.4811424,
                                   44.4840562, 44.4869461]
    assert list(leap_dates) == [2441499.5, 2441683.5, 2442048.5]
    assert list(leap_offsets) == [11, 12, 13]

    ts = Timescale((daily_tt, daily_delta_t), leap_dates, leap_offsets)
    t = ts.tai(1973, 1, 1, 0, 0, [11, 12])
    assert t.utc.T.tolist() == [
        [1972, 12, 31, 23, 59, 60],
        [1973, 1, 1, 0, 0, 0],
    ]

def test_build_timescale_arrays_when_series_already_has_early_leap_seconds():
    mjd = array([41497.0, 41498.0, 41499.0, 41500.0])
    dut1 = array([-0.6324066, -0.6349935, 0.3621956, 0.3592006])
    daily_tt, daily_delta_t, leap_dates, leap_offsets = (
        iers.build_timescale_arrays(mjd, dut1))

    assert list(leap_dates) == [2441499.5]
    assert list(leap_offsets) == [11.0]



"""Parse data files from the International Earth Rotation Service.

See:
https://datacenter.iers.org/eop.php
ftp://cddis.gsfc.nasa.gov/pub/products/iers/readme.finals2000A

"""
import numpy as np
from ..constants import DAY_S

inf = float('inf')

# This regular expression must remain a plain string; attempting to
# compile it triggers a bug in older NumPy versions like 1.14.3:
# https://github.com/skyfielders/python-skyfield/issues/372
_R = (b'(?m)^......(.........) . '
      b'(.\\d.......)......... '
      b'(.\\d.......).........  '
      b'.(.\\d........)')

def parse_x_y_dut1_from_finals_all(f):
    return np.fromregex(f, _R, [
        ('utc_mjd', float),
        ('x_arcseconds', float),
        ('y_arcseconds', float),
        ('dut1', float),
    ])

def install_polar_motion_table(ts, finals_data):
    t = ts.utc(1858, 11, 17.0 + finals_data['utc_mjd'])
    ts.polar_motion_table = (
        t.tt,
        np.array(finals_data['x_arcseconds']),
        np.array(finals_data['y_arcseconds']),
    )

def build_timescale_arrays(utc_mjd, dut1):
    big_jumps = np.diff(dut1) > 0.9
    leap_second_mask = np.concatenate([[False], big_jumps])
    tt_minus_utc = np.cumsum(leap_second_mask) + 32.184 + 12.0

    daily_tt = utc_mjd + tt_minus_utc / DAY_S + 2400000.5
    daily_delta_t = (tt_minus_utc - dut1).round(7)
    leap_dates = utc_mjd[leap_second_mask]

    # Since "finals2000A.all" starts on 1973-01-02 and leaves out the
    # first two leap seconds, add them back.  (But check first, in case
    # the user loads a more complete data file than "finals2000A.all".)
    first_leap = leap_dates[0] if len(leap_dates) else 0
    more_leaps = [mjd for mjd in (41499.0, 41683.0) if first_leap > mjd]
    leap_dates = np.concatenate([more_leaps, leap_dates])

    leap_dates += 2400000.5
    leap_offsets = np.arange(11.0, len(leap_dates) + 11.0)
    return daily_tt, daily_delta_t, leap_dates, leap_offsets

# Compatibility with older Skyfield versions:

def parse_dut1_from_finals_all(f):
    data = parse_x_y_dut1_from_finals_all(f)
    return data['utc_mjd'], data['dut1']



from __future__ import division

import sys
import math
from numpy import (
    abs, amax, amin, arange, arccos, arctan, array, atleast_1d,
    clip, copy, copyto, cos, cosh, exp, full_like, log, ndarray, newaxis,
    pi, power, repeat, sign, sin, sinh, sqrt, sum, tan, tanh, zeros_like,
)
from skyfield.constants import AU_KM, DAY_S, DEG2RAD
from skyfield.functions import dots, length_of, mxv
from skyfield.descriptorlib import reify
from skyfield.elementslib import OsculatingElements, normpi
from skyfield.units import Distance, Velocity
from skyfield.vectorlib import VectorFunction
from skyfield.sgp4lib import _cross

_CONVERT_GM = DAY_S * DAY_S / AU_KM / AU_KM / AU_KM

class _KeplerOrbit(VectorFunction):
    def __init__(self,
                 position,
                 velocity,
                 epoch,
                 mu_au3_d2,
                 center=None,
                 target_name=None,
        ):
        """ Calculates the position of an object using 2 body propagation

        Parameters
        ----------
        position : Distance
            Position vector at epoch with shape (3,)
        velocity : Velocity
            Velocity vector at epoch with shape (3,)
        epoch : Time
            Time corresponding to `position` and `velocity`
        mu_au_d : float
            Value of mu (G * M) in au^3/d^2
        center : int
            NAIF ID of the primary body, 399 for geocentric orbits, 10 for
            heliocentric orbits
        target : int
            NAIF ID of the secondary body
        """
        self.position_at_epoch = position
        self.velocity_at_epoch = velocity
        self.epoch = epoch
        self.mu_au3_d2 = mu_au3_d2
        self.center = center
        self.target_name = target_name

        self._rotation = None  # TODO: make argument?

    @property
    def target(self):
        return self  # this instance itself represents the target object

    @classmethod
    def _from_periapsis(
            cls,
            semilatus_rectum_au,
            eccentricity,
            inclination_degrees,
            longitude_of_ascending_node_degrees,
            argument_of_perihelion_degrees,
            t_periapsis,
            gm_km3_s2,
            center=None,
            target_name=None,
        ):
        """Build a `KeplerOrbit` given its parameters and date of periapsis."""
        gm_au3_d2 = gm_km3_s2 * _CONVERT_GM
        pos, vel = ele_to_vec(
            semilatus_rectum_au,
            eccentricity,
            DEG2RAD * inclination_degrees,
            DEG2RAD * longitude_of_ascending_node_degrees,
            DEG2RAD * argument_of_perihelion_degrees,
            0.0,
            gm_au3_d2,
        )
        return cls(
            Distance(pos),
            Velocity(vel),
            t_periapsis,
            gm_au3_d2,
            center,
            target_name,
        )

    @classmethod
    def _from_true_anomaly(cls, p, e, i, Om, w, v,
                          epoch,
                          mu_km_s=None,
                          mu_au3_d2=None,
                          center=None,
                          target_name=None,
        ):
        """ Creates a `KeplerOrbit` object from elements using true anomaly

        Parameters
        ----------
        p : Distance
            Semi-Latus Rectum
        e : float
             Eccentricity
        i : Angle
            Inclination
        Om : Angle
            Longitude of Ascending Node
        w : Angle
            Argument of periapsis
        v : Angle
            True anomaly
        epoch : Time
            Time corresponding to `position` and `velocity`
        mu_km_s : float
            Value of mu (G * M) in km^3/s^2
        mu_au3_d2 : float
            Value of mu (G * M) in au^3/d^2
        center : int
            NAIF ID of the primary body, 399 for geocentric orbits, 10 for
            heliocentric orbits
        target : int
            NAIF ID of the secondary body
        """
        if (mu_km_s and mu_au3_d2) or (not mu_km_s and not mu_au3_d2):
            raise ValueError('Either mu_km_s or mu_au3_d2 should be used, but not both')

        if mu_au3_d2:
            mu_km_s = mu_au3_d2 * AU_KM**3 / DAY_S**2

        position, velocity = ele_to_vec(p.km,
                                        e,
                                        i.radians,
                                        Om.radians,
                                        w.radians,
                                        v.radians,
                                        mu_km_s,
        )
        return cls(Distance(km=position),
                   Velocity(km_per_s=velocity),
                   epoch,
                   mu_km_s,
                   center=center,
                   target_name=target_name,
        )

    @classmethod
    def _from_mean_anomaly(
            cls,
            semilatus_rectum_au,
            eccentricity,
            inclination_degrees,
            longitude_of_ascending_node_degrees,
            argument_of_perihelion_degrees,
            mean_anomaly_degrees,
            epoch,
            gm_km3_s2,
            center=None,
            target_name=None,
        ):
        """ Creates a `KeplerOrbit` object from elements using mean anomaly

        Parameters
        ----------
        p : Distance
            Semi-Latus Rectum
        e : float
             Eccentricity
        i : Angle
            Inclination
        Om : Angle
            Longitude of Ascending Node
        w : Angle
            Argument of periapsis
        M : Angle
            Mean anomaly
        epoch : Time
            Time corresponding to `position` and `velocity`
        mu_km_s : float
            Value of mu (G * M) in km^3/s^2
        mu_au3_d2 : float
            Value of mu (G * M) in au^3/d^2
        center : int
            NAIF ID of the primary body, 399 for geocentric orbits, 10 for
            heliocentric orbits
        target : int
            NAIF ID of the secondary body
        """
        M = DEG2RAD * mean_anomaly_degrees
        gm_au3_d2 = gm_km3_s2 * _CONVERT_GM
        if eccentricity < 1.0:
            E = eccentric_anomaly(eccentricity, M)
            v = true_anomaly_closed(eccentricity, E)
        elif eccentricity > 1.0:
            E = eccentric_anomaly(eccentricity, M)
            v = true_anomaly_hyperbolic(eccentricity, E)
        else:
            v = true_anomaly_parabolic(semilatus_rectum_au, gm_au3_d2, M)

        pos, vel = ele_to_vec(
            semilatus_rectum_au,
            eccentricity,
            DEG2RAD * inclination_degrees,
            DEG2RAD * longitude_of_ascending_node_degrees,
            DEG2RAD * argument_of_perihelion_degrees,
            v,
            gm_au3_d2,
        )
        return cls(
            Distance(pos),
            Velocity(vel),
            epoch,
            gm_au3_d2,
            center,
            target_name,
        )

    def _at(self, time):
        """Propagate the KeplerOrbit to the given Time object

        The Time object can contain one time, or an array of times
        """
        pos, vel = propagate(
            self.position_at_epoch.au,
            self.velocity_at_epoch.au_per_d,
            self.epoch.tt,
            time.tt,
            self.mu_au3_d2,
        )
        if self._rotation is not None:
            pos = mxv(self._rotation, pos)
            vel = mxv(self._rotation, vel)
        return pos, vel, None, None

    @reify
    def elements_at_epoch(self):
        return OsculatingElements(self.position_at_epoch,
                                  self.velocity_at_epoch,
                                  self.epoch,
                                  mu_km_s = self.mu_au3_d2 / _CONVERT_GM,
        )

    def __str__(self):
        ele = self.elements_at_epoch
        if self.target_name:
            return 'KeplerOrbit {0} -> {1}'.format(
                self.center_name,
                self.target_name,
            )
        ele = self.elements_at_epoch
        string = ('KeplerOrbit {} ->'
                  ' q={:.2}au e={:.3f} i={:.1f} Om={:.1f} w={:.1f}')
        return string.format(
            self.center_name,
            ele.periapsis_distance.au,
            ele.eccentricity,
            ele.inclination.degrees,
            ele.longitude_of_ascending_node.degrees,
            ele.argument_of_periapsis.degrees,
        )

    def __repr__(self):
        return '<{0}>'.format(str(self))

_ten_iterations = tuple([None] * 10)

def eccentric_anomaly(e, M):
    """Iterate to solve Kepler's equation to find the eccentric anomaly.

    See arXiv:2108.03215.

    """
    M = normpi(M)
    sign_M = sign(M)
    M *= sign_M
    ebar = 0.25 * pi/e - 1.0
    E = 0.5 * pi * ebar * (sign(ebar) * sqrt(1 + M/(e*ebar*ebar)) - 1.0)

    for _ in _ten_iterations:
        f1 = 1.0 - e*cos(E)
        f2 = e*sin(E)
        f = E - f2 - M
        dE = f*f1 / (f1*f1 - 0.5*f*f2)
        E -= dE
        if abs(dE) < 1e-14:
            return E * sign_M

    raise ValueError('eccentric anomaly failed to converge')

def true_anomaly_hyperbolic(e, E):
    """Calculates true anomaly from eccentricity and eccentric anomaly.

    Valid for hyperbolic orbits. Equations from the relevant Wikipedia entries.

    """
    return 2.0 * arctan(sqrt((e + 1.0) / (e - 1.0)) * tanh(E/2))

def true_anomaly_closed(e, E):
    """Calculates true anomaly from eccentricity and eccentric anomaly.

    Valid for closed orbits. Equations from the relevant Wikipedia entries.

    """
    return 2.0 * arctan(sqrt((1.0 + e) / (1.0 - e)) * tan(E/2))

def true_anomaly_parabolic(p, gm, M):
    """Calculates true anomaly from semi-latus rectum, gm, and mean anomaly.

    Valid for parabolic orbits. Equations from
    https://en.wikipedia.org/wiki/Parabolic_trajectory.

    """
    delta_t = sqrt(2 * p**3 / gm) * M # from http://www.bogan.ca/orbits/kepler/orbteqtn.html
    periapsis_distance = p / 2
    A = 3 / 2 * sqrt(gm / (2 * periapsis_distance**3)) * delta_t
    B = (A + (A*A + 1))**(1/3)
    return 2 * arctan(B - 1/B)

def ele_to_vec(p, e, i, Om, w, v, mu):
    """Calculates state vectors from orbital elements. Also checks for invalid
    sets of elements.

    Based on equations from this document:

    https://web.archive.org/web/*/http://ccar.colorado.edu/asen5070/handouts/kep2cart_2002.doc
    """
    # Checks that true anomaly is less than arccos(-1/e) for hyperbolic orbits
    if isinstance(e, ndarray) and isinstance(v, ndarray):
        inds = (e>1)
        if (v[inds]>arccos(-1/e[inds])).any():
            raise ValueError('If eccentricity is >1, abs(true anomaly) cannot be more than arccos(-1/e)')
    elif isinstance(e, ndarray) and not isinstance(v, ndarray):
        inds = (e>1)
        if (v>arccos(-1/e[inds])).any():
            raise ValueError('If eccentricity is >1, abs(true anomaly) cannot be more than arccos(-1/e)')
    elif isinstance(v, ndarray) and not isinstance(e, ndarray):
        if e>1 and (v>arccos(-1/e)).any():
            raise ValueError('If eccentricity is >1, abs(true anomaly) cannot be more than arccos(-1/e)')
    else:
        if e>1 and v>arccos(-1/e):
            raise ValueError('If eccentricity is >1, abs(true anomaly) cannot be more than arccos(-1/e)')

    # Checks that inclination is in the range [0, pi]
    if isinstance(i, ndarray):
        if not ((i>=0) * (i <= pi)).all():
            raise ValueError('Inclination outside the range [0, pi] radians')
    else:
        if not 0 <= i <= pi:
            raise ValueError('Inclination outside the range [0, pi] radians')

    r = p/(1 + e*cos(v))
    h = sqrt(p*mu)
    u = v+w

    X = r*(cos(Om)*cos(u) - sin(Om)*sin(u)*cos(i))
    Y = r*(sin(Om)*cos(u) + cos(Om)*sin(u)*cos(i))
    Z = r*(sin(i)*sin(u))

    X_dot = X*h*e/(r*p)*sin(v) - h/r*(cos(Om)*sin(u) + sin(Om)*cos(u)*cos(i))
    Y_dot = Y*h*e/(r*p)*sin(v) - h/r*(sin(Om)*sin(u) - cos(Om)*cos(u)*cos(i))
    Z_dot = Z*h*e/(r*p)*sin(v) + h/r*sin(i)*cos(u)

    # z and z_dot are independent of Om, so if Om is an array and the other
    # elements are scalars, z and z_dot need to be repeated
    if Z.size!=X.size:
        Z = repeat(Z, X.size)
        Z_dot = repeat(Z_dot, X.size)

    return array([X, Y, Z]), array([X_dot, Y_dot, Z_dot])

dpmax = sys.float_info.max

def find_trunc():
    denom = 2
    factr = 2
    trunc = 1
    x = 1 / denom
    while 1+x > 1:
        denom = denom * (2+factr) * (1+factr)
        factr = factr + 2
        trunc = trunc + 1
        x = 1 / denom
    return trunc

trunc = find_trunc()
odd_factorials = array([math.factorial(i) for i in range(3, trunc*2, 2)])
even_factorials = array([math.factorial(i) for i in range(2, trunc*2, 2)])
exponents = arange(0, trunc-1)
stumpff_bound = -(log(2) + log(dpmax))**2

def stumpff(x):
    """Calculates Stumpff functions

    Based on the function toolkit/src/spicelib/stmp03.f from the SPICE toolkit,
    which can be downloaded from naif.jpl.nasa.gov/naif/toolkit_FORTRAN.html
    """
    if x.min() < stumpff_bound:
        raise ValueError('Argument below lower bound')

    z = sqrt(abs(x))

    c0 = zeros_like(x)
    c1 = zeros_like(x)
    c2 = zeros_like(x)
    c3 = zeros_like(x)

    low = x < -1
    c0[low] = cosh(z[low])
    c1[low] = sinh(z[low])/z[low]

    high = x > 1
    c0[high] = cos(z[high])
    c1[high] = sin(z[high])/z[high]

    mid = ~(low|high)
    if sum(mid):
        numerators = repeat(x[mid][:, newaxis], trunc-1, axis=1)
        numerators[:, 1::2] *= -1
        c3[mid] = sum(power(numerators, exponents)/odd_factorials, axis=1)
        c2[mid] = sum(power(numerators, exponents)/even_factorials, axis=1)

        c1[mid] = 1 - x[mid]*c3[mid]
        c0[mid] = 1 - x[mid]*c2[mid]

    not_mid = ~mid
    c2[not_mid] = (1 - c0[not_mid])/x[not_mid]
    c3[not_mid] = (1 - c1[not_mid])/x[not_mid]

    return c0, c1, c2, c3

def propagate(position, velocity, t0, t1, gm):
    """Propagates a position and velocity vector with an array of times.

    Based on the function toolkit/src/spicelib/prop2b.f from the SPICE toolkit,
    which can be downloaded from naif.jpl.nasa.gov/naif/toolkit_FORTRAN.html

    Parameters
    ----------
    position : ndarray
       Position vector with shape (3,)
    velocity : ndarray
        Velocity vector with shape (3,)
    t0 : float
        Time corresponding to `position` and `velocity`
    t1 : float or ndarray
        Time or times to propagate to
    gm : float
        Gravitational parameter in units that match the other arguments
    """
    output_shape = (3,) + t1.shape

    gm = atleast_1d(gm)
    if (gm <= 0).any():
        raise ValueError("'gm' should be positive")
    if (length_of(velocity)).any() == 0:
        raise ValueError('Velocity vector has zero magnitude')
    if (length_of(position)).any() == 0:
        raise ValueError('Position vector has zero magnitude')

    if position.ndim == 1:
        position = position[:, newaxis]
    if velocity.ndim == 1:
        velocity = velocity[:, newaxis]

    r0 = length_of(position)
    rv = dots(position, velocity)

    hvec = _cross(position, velocity)
    h2 = dots(hvec, hvec)

    if (h2 == 0).any():
        raise ValueError('Motion is not conical')

    eqvec = _cross(velocity, hvec)/gm + -position/r0
    e = length_of(eqvec)
    q = h2 / (gm * (1+e))

    f = 1 - e
    b = sqrt(q/gm)

    br0 = b * r0
    b2rv = b * b * rv
    bq = b * q
    qovr0 = q / r0

    maxc = amax(array([abs(br0),
               abs(b2rv),
               abs(bq),
               abs(qovr0/bq)]), axis=0)

    hyperbolic = (f<0)
    bound = zeros_like(f)

    fixed = log(dpmax/2) - log(maxc[hyperbolic])
    rootf = sqrt(-f[hyperbolic])
    logf = log(-f[hyperbolic])
    bound[hyperbolic] = amin(array([fixed/rootf, (fixed + 1.5*logf)/rootf]), axis=0)

    logbound = (log(1.5) + log(dpmax) - log(maxc[~hyperbolic])) / 3
    bound[~hyperbolic] = exp(logbound)

    # each of these arrays has 1 entry per orbit, so its shape is (#orbits, 1)
    f = f[:, newaxis]
    bq = bq[:, newaxis]
    b2rv = b2rv[:, newaxis]
    br0 = br0[:, newaxis]
    qovr0 = qovr0[:, newaxis]
    bound = bound[:, newaxis]

    def kepler(x):
        _, c1, c2, c3 = stumpff(f*x*x)
        return x*(br0*c1 + x*(b2rv*c2 + x*bq*c3))

    def kepler_1d(x, orb_inds):
        _, c1, c2, c3 = stumpff(x*x*repeat(f, orb_inds))
        return x*(c1*repeat(br0, orb_inds) + x*(c2*repeat(b2rv, orb_inds) + x*(c3*repeat(bq, orb_inds))))

    t1 = atleast_1d(t1)
    t0 = atleast_1d(t0)
    if len(t0) == 1:
        t0 = repeat(t0, position.shape[1])

    # shape of 2 dimensional arrays from here on out should be (#orbits, len(t1))
    dt = t1 - t0[:, newaxis]

    x = dt/bq
    copyto(x, -bound, where=(x<-bound))
    copyto(x, bound, where=(x>bound))

    kfun = kepler(x)

    past = dt < 0
    future = dt > 0

    upper = zeros_like(dt, dtype='float64')
    lower = zeros_like(dt, dtype='float64')
    oldx = zeros_like(dt, dtype='float64')

    copyto(lower, x, where=past)
    copyto(upper, x, where=future)

    while (kfun[past] > dt[past]).any():
        copyto(upper, lower, where=past)
        lower[past] *= 2
        copyto(oldx, x, where=past)
        orb_ind = sum(past, axis=1)
        x[past] = clip(lower[past], repeat(-bound, orb_ind), repeat(bound, orb_ind))
        if (x[past] == oldx[past]).any():
            raise ValueError('The input delta time (dt) has a value of {0}.'
                             'This is beyond the range of DT for which we '
                             'can reliably propagate states. The limits for '
                             'this GM and initial state are from {1}'
                             'to {2}.'.format(dt, kepler(-bound), kepler(bound)))
        kfun[past] = kepler_1d(x[past], orb_ind)

    while (kfun[future] < dt[future]).any():
        copyto(lower, upper, where=future)
        upper[future] *= 2
        copyto(oldx, x, where=future)
        orb_ind = sum(future, axis=1)
        x[future] = clip(upper[future], repeat(-bound, orb_ind), repeat(bound, orb_ind))
        if (x[future] == oldx[future]).any():
            raise ValueError('The input delta time (dt) has a value of {0}.'
                             'This is beyond the range of DT for which we '
                             'can reliably propagate states. The limits for '
                             'this GM and initial state are from {1} '
                             'to {2}.'.format(dt, kepler(-bound), kepler(bound)))
        kfun[future] = kepler_1d(x[future], orb_ind)

    x = copy(upper)
    copyto(x, (upper+lower)/2, where=(lower<=upper))

    lcount = zeros_like(dt)
    mostc = full_like(dt, 1000)
    not_done = (lower < x) & (x < upper)

    while not_done.any():
        orb_inds = sum(not_done, axis=1)
        kfun[not_done] = kepler_1d(x[not_done], orb_inds)

        high = (kfun > dt) & not_done
        low = (kfun < dt) & not_done
        same = (~high & ~low) & not_done

        copyto(upper, x, where=(high|same))
        copyto(lower, x, where=(low|same))

        condition = not_done & (mostc > 64) & (upper != 0) & (lower != 0)
        mostc[condition] = 64
        lcount[condition] = 0

        copyto(x, upper, where=(not_done & (lower>upper)))
        copyto(x, (upper+lower)/2, where=(not_done & (lower<=upper)))

        lcount += 1
        not_done = (lower < x) & (x < upper) & (lcount < mostc)

    c0, c1, c2, c3 = stumpff(f*x*x)
    br = br0*c0 + x*(b2rv*c1 + x*bq*c2)

    pc = 1 - qovr0 * x * x * c2
    vc = dt - bq * x**3 * c3
    pcdot = -qovr0 / br * x * c1
    vcdot = 1 - bq / br * x * x * c2

    position_prop = pc[newaxis, :, :]*position[:, :, newaxis] + vc[newaxis, :, :]*velocity[:, :, newaxis]
    velocity_prop = pcdot[newaxis, :, :]*position[:, :, newaxis] + vcdot[newaxis, :, :]*velocity[:, :, newaxis]

    position_prop.shape = output_shape
    velocity_prop.shape = output_shape
    return position_prop, velocity_prop



from jplephem.names import (
    target_name_pairs as code_name_pairs,
    target_names as code_names
)

name_codes = dict((name, code) for code, name in code_name_pairs)

def numbered_name_of(code):
    """Given a code, return a string giving both the code and name.

    >>> numbered_name_of(301)
    '301 Moon'

    """
    name = code_names.get(code, '(Unnamed)')
    return '{0} {1}'.format(code, name)

def _target_name(target):
    """Return `target` annotated for display to the user.

    * A string target is quoted to make clear that it's a string, like 'comet'.
    * A numeric target has its NAIF name appended, like 399 EARTH.

    """
    if isinstance(target, str):
        return repr(target)
    name = code_names.get(target, 'UNKNOWN')
    return '{0} {1}'.format(target, name)



from numpy import abs, clip, einsum, sqrt, where

from .constants import C, AU_M, C_AUDAY, GS
from .functions import _AVOID_DIVIDE_BY_ZERO, dots, length_of

deflectors = ['sun', 'jupiter', 'saturn', 'moon', 'venus', 'uranus', 'neptune']
rmasses = {
    # earth-moon barycenter: 328900.561400
    199: 6023600.0,             # mercury
    299: 408523.71,             # venus
    399: 332946.050895,         # earth
    499: 3098708.0,             # mars
    599: 1047.3486,             # jupiter
    699: 3497.898,              # saturn
    799: 22902.98,              # uranus
    899: 19412.24,              # neptune
    999: 135200000.0,           # pluto
    10: 1.0,                    # sun
    301: 27068700.387534,       # moon

    # For compatibility with any user code that discovered and used this
    # undocumented `rmasses` dict prior to Skyfield 1.53:
    'mercury': 6023600.0,
    'venus': 408523.71,
    'earth': 332946.050895,
    'mars': 3098708.0,
    'jupiter': 1047.3486,
    'saturn': 3497.898,
    'uranus': 22902.98,
    'neptune': 19412.24,
    'pluto': 135200000.0,
    'sun': 1.0,
    'moon': 27068700.387534,
}

def add_deflection(position, observer, ephemeris, t,
                   include_earth_deflection, count=3):
    """Update `position` for how solar system masses will deflect its light.

    Given the ICRS `position` |xyz| of an object (au) that is being
    viewed from the `observer` also expressed as |xyz|, and given an
    ephemeris that can be used to determine solar system body positions,
    and given the time `t` and Boolean `apply_earth` indicating whether
    to worry about the effect of Earth's mass, and a `count` of how many
    major solar system bodies to worry about, this function updates
    `position` in-place to show how the masses in the solar system will
    deflect its image.

    """
    # Compute light-time to observed object.

    tlt = length_of(position) / C_AUDAY

    # Cycle through gravitating bodies.

    for name in deflectors[:count]:
        try:
            deflector = ephemeris[name]
        except KeyError:
            deflector = ephemeris[name + ' barycenter']

        rmass = rmasses[name]
        pe = _compute_deflector_position(
            t, observer, position, deflector, tlt,
        )
        position += _compute_deflection(position, pe, rmass)

    # If observer is not at geocenter, add in deflection due to Earth.

    if include_earth_deflection.any():
        deflector = ephemeris['earth']
        bposition = deflector.at(t).xyz.au
        rmass = rmasses['earth']
        pe = observer - bposition
        d = _compute_deflection(position, pe, rmass)
        if include_earth_deflection.shape:
            d *= include_earth_deflection  # where False, set `d` to zero
        position += d

def _compute_deflector_position(t, observer, position, deflector, tlt):
    """Compute where a deflector was when closest to a light beam."""

    # Get position of gravitating body wrt observer.

    bposition = deflector.at(t).xyz.au
    gpv = bposition - observer

    # Compute light-time from point on incoming light ray that is closest
    # to gravitating body.

    dlt = light_time_difference(position, gpv)

    # Should really add TDB offset, not TT, but doesn't matter.

    tclose = t - clip(dlt, 0.0, tlt)

    # Get position of gravitating body wrt ss barycenter at time when
    # incoming photons were closest to it.

    bposition = deflector.at(tclose).xyz.au
    pe = observer - bposition
    return pe

def light_time_difference(position, deflector_position):
    """When did the light from ``position`` pass closest to a given deflector?

    Given an observer at the origin and a ``position`` |xyz| in AU, how
    recently did the light from the position pass closest to the object
    whose |xyz| is given as ``deflector_position``?  The answer is
    returned as a floating point number of days.

    """
    # From 'pos1', form unit vector 'u1' in direction of star or light
    # source.

    dis = length_of(position)
    u1 = position / (dis + _AVOID_DIVIDE_BY_ZERO)

    # Light-time returned is the projection of vector 'pos_obs' onto the
    # unit vector 'u1' (formed from 'pos1'), divided by the speed of light.

    diflt = einsum('a...,a...', u1, deflector_position) / C_AUDAY
    return diflt

def _compute_deflection(position, pe, rmass):
    """Compute how much a mass will deflect a position.

    The ``position`` is relative to the observer in AU; ``pe`` is the
    position of the observer relative to the deflector in AU; and
    ``rmass`` is the deflector's reciprocal mass.

    """
    pq = position + pe

    # Compute vector magnitudes and unit vectors.

    pmag = length_of(position)
    qmag = length_of(pq)
    emag = length_of(pe)

    phat = position / where(pmag, pmag, 1.0)  # where() avoids divide-by-zero
    qhat = pq / where(qmag, qmag, 1.0)
    ehat = pe / where(emag, emag, 1.0)

    # Compute dot products of vectors.

    pdotq = dots(phat, qhat)
    qdote = dots(qhat, ehat)
    edotp = dots(ehat, phat)

    # If gravitating body is observed object, or is on a straight line
    # toward or away from observed object to within 1 arcsec, deflection
    # is set to zero.

    flag = abs(edotp) <= 0.99999999999

    # Compute scalar factors.

    fac1 = 2.0 * GS / (C * C * emag * AU_M * rmass)
    fac2 = 1.0 + qdote

    return flag * fac1 * (pdotq * ehat - edotp * qhat) / fac2 * pmag

def add_aberration(position, velocity, light_time):
    """Correct a relative position vector for aberration of light.

    Given the relative `position` |xyz| of an object (AU) from a
    particular observer, the `velocity` [dx,dy,dz] at which the observer
    is traveling (AU/day), and the light propagation delay `light_time`
    to the object (days), this function updates `position` in-place to
    give the object's apparent position due to the aberration of light.

    """
    p1mag = light_time * C_AUDAY
    vemag = length_of(velocity)
    beta = vemag / C_AUDAY
    dot = dots(position, velocity)

    cosd = dot / (p1mag * vemag + _AVOID_DIVIDE_BY_ZERO)
    gammai = sqrt(1.0 - beta * beta)
    p = beta * cosd
    q = (1.0 + p / (1.0 + gammai)) * light_time
    r = 1.0 + p

    position *= gammai
    position += q * velocity
    position /= r



# -*- coding: utf-8 -*-
"""Parsing routines for JPL text PCK files.

For an (incomplete) summary of the file format, look for the heading
“NAIF Text Kernel Format” in the “PCK Required Reading”:

https://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/pck.html

"""
import re

def load(lines, variables):
    """Load PCK text kernel names and values into a ``variables`` dict."""
    for name, equals, values in parse(lines):
        if equals == b'=':
            if len(values) == 1:
                variables[name] = values[0]  # unbox a scalar value
            else:
                variables[name] = values
        elif equals == b'+=':
            old = variables.get(name)
            if old is None:
                old = variables[name] = []
            elif not isinstance(old, list):
                old = [old]  # box an earlier scalar before extending
            old.extend(values)

def parse(lines):
    """Yield ``(name, equals, values)`` bytestrings from a PCK text kernel.

    This merely reads raw assignment statements; it doesn’t combine
    multiple ``+=`` assignments to create single values.  The byte
    string ``equals`` will be either ``b'='`` or ``b'+='``.  Scalars are
    returned as a ``values`` list one item long.

    """
    tokens = iter(_parse_tokens(lines))
    for token in tokens:
        name = token.decode('ascii')
        equals = next(tokens)
        if equals not in (b'=', b'+='):
            raise ValueError('an equals sign is expected after %r' % name)
        token = next(tokens)
        if token == b'(':
            values = []
            for token in tokens:
                if token == b')':
                    break
                values.append(_evaluate(token))
        else:
            values = [_evaluate(token)]
        yield name, equals, values

def _evaluate(token):
    """Return a string, integer, or float parsed from a PCK text kernel."""
    if token[0:1].startswith(b"'"):
        return token[1:-1].decode('ascii')
    if token.isdigit():
        return int(token)
    if token.startswith(b'@'):
        raise NotImplementedError('TODO: need parser for dates,'
                                  ' like @01-MAY-1991/16:25')
    token = token.replace(b'D', b'E')  # for numbers like -1.4D-12
    return float(token)

_token_re = re.compile(b"[A-Za-z]\\w+|=|\\+=|\\(|\\)|'[^']*'|[^), ]+")

def _parse_tokens(lines):
    """Yield all the tokens inside the data segments of a PCK text file."""
    lines = iter(lines)
    for line in lines:
        if b'\\begindata' not in line:
            continue  # save cost of strip() on most lines
        line = line.strip()
        if line != b'\\begindata':
            continue
        for line in lines:
            line = line.strip()
            if line == b'\\begintext':
                break
            for token in _token_re.findall(line):
                yield token
